# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Extensions such as `jit`, `grad`, `logsumexp`, etc."""
import bisect
import contextlib
import copy
import functools
import string
import sys
import threading
import numpy as np
import six

import tensorflow.compat.v2 as tf

import trax.tf_numpy.numpy as tf_np

_int_dtype_lower_bounds = [
    -2**63, -2**31, -2**15, -2**7, 0, 2**7, 2**15, 2**31, 2**64
]
_int_dtypes = [
    tf.int64, tf.int32, tf.int16, tf.int8, tf.uint8, tf.uint16, tf.uint32,
    tf.uint64
]
_tf_nn_APIs = {1: [tf.nn.conv1d, tf.nn.conv1d_transpose],
               2: [tf.nn.conv2d, tf.nn.conv2d_transpose],
               3: [tf.nn.conv3d, tf.nn.conv3d_transpose]}


remat = tf.recompute_grad


def most_precise_int_dtype(x):
  if not isinstance(x, six.integer_types) or isinstance(x, bool):
    return None
  i = bisect.bisect_right(_int_dtype_lower_bounds, x)
  if i in (0, len(_int_dtype_lower_bounds)):
    raise ValueError("Integer %s is out of bounds" % x)
  assert len(_int_dtype_lower_bounds) == len(_int_dtypes) + 1
  return _int_dtypes[i - 1]


def _canonicalize_jit_arg(x):
  if isinstance(x, tf_np.ndarray):
    return x
  try:
    # We need to convert `int` to the most precise dtype, otherwise the dtype
    # of the result may be different from numpy's. For example, when a binary
    # op takes in a Python integer 5 and an array of uint32, numpy will pick
    # uint32 as 5's dtype, while tf.convert_to_tensor will choose int32 which
    # will cause the two arguments to be promoted to int64. We pick uint8
    # here, which will be promoted to uint32 by the binary op.
    # Note that we prefer unsigned int to signed int when both are equally
    # precise. For example, for 5, we pick uint8 instead of int8. There is no
    # reason to prefer one to the other, because for each there is a case
    # where the behavior diverges from numpy. If we prefer signed int,
    # consider the case where the first operand is 5 and the second is
    # 2**64-1. Numpy picks uint64 as the result dtype, but because we choose a
    # signed type for 5 such as int8, the result type will be float64. On the
    # other hand, if we prefer unsigned int, consider the case where the first
    # operand is 2**31-1 and the second is -1. Numpy will pick int32, but
    # because we choose uint32 for 2*32-1, the result will be int64. The root
    # of the problem is that `jit` converts `int` to tensors (hence committing
    # to a dtype) too early, when we don't have enough information about the
    # jitted function (e.g. which subset of the arguments should be promoted
    # together using np.result_type). tf.function doesn't have this problem
    # because it doesn't convert `int` to tensors. jax.jit doesn't have this
    # problem because it converts `int` to "int tracer" which doesn't commit
    # to a dtype.
    # TODO(wangpeng): Revisit this design and see whether we can improve `jit`
    #   and tf.function.
    dtype = most_precise_int_dtype(x)
    if dtype is None and isinstance(x, float):
      dtype = tf_np.default_float_type()
    return tf.convert_to_tensor(value=x, dtype=dtype)
  except (TypeError, ValueError):
    return x


def _canonicalize_jit_arguments(inp):
  """Canonicalize arguments to be used for jit.

  Args:
    inp: a nested structure of arguments to be canonicalized (i.e. to be
      converted to Tensors). Only tf_np.ndarray and things accepted by
      `tf.convert_to_tensor` will be converted.

  Returns:
    The canonicalized version.
  """
  return tf.nest.map_structure(_canonicalize_jit_arg, inp)


def _tf_to_np(inp):

  def f(x):
    if isinstance(x, tf.IndexedSlices):
      return tf_np.asarray(x)
    else:
      return x

  return tf.nest.map_structure(f, inp)


def stop_gradient(x):

  def static_stop_gradient(x):
    # `tf.stop_gradient` is a no-op for non-Tensor. Returning the original type
    # allows it to be used in the conditional without Autograph, if static. For
    # example:
    # `if fastmath.stop_gradient(5) > 4:`
    return tf.stop_gradient(x) if tf.is_tensor(x) else x

  return _tf_to_np(tf.nest.map_structure(static_stop_gradient, x))


def custom_grad(f_vjp, f_original=None):
  """Decorator to define a function with a custom gradient.

  This function is very similar to `tf.custom_gradient`. See the documentation
  of `tf.custom_gradient` for detailed usage.

  The differences with `tf.custom_gradient` are:

  - All arguments and results are tf_np.ndarrays instead of tensors.

  - The `grad_fn` returned by `f_vjp` accepts and returns nested structures,
    unlike that in `tf.custom_gradient` which only accepts and returns lists.

  Args:
    f_vjp: the same as the `f` argument of `tf.custom_gradient`. Note that all
      inputs and outputs of `f_vjp` and of the `grad_fn` function it returns can
      be nested structures.
    f_original: (optional) not used.

  Returns:
    The same as `tf.custom_gradient`.
  """
  del f_original

  @tf.custom_gradient
  def tf_f(*tf_args, **tf_kwargs):
    np_args = _tf_to_np(tf_args)
    np_kwargs = _tf_to_np(tf_kwargs)
    np_y, np_vjp = f_vjp(*np_args, **np_kwargs)
    tf_y = np_y

    def tf_vjp(*flat_tf_dy):
      tf_dy = tf.nest.pack_sequence_as(tf_y, flat_tf_dy)
      np_dy = _tf_to_np(tf_dy)
      np_dx = np_vjp(np_dy)
      return tf.nest.flatten(np_dx)

    return tf_y, tf_vjp

  def np_f(*args, **kwargs):
    return _tf_to_np(tf_f(*args), **kwargs)

  return np_f


def vjp(f, *primals, has_aux=False):
  """Returns the result and the VJP function of `f`.

  This function returns the result and the vector-Jacobian-product (VJP)
  function of `f`.

  Args:
    f: a function from (nested structures of) tf_np.ndarrays to a (nested
      structure of) tf_np.ndarray. If `has_aux` is True, it should return an
      extra output.
    *primals: the inputs to be fed to `f`.
    has_aux: if True, the second output of `f` will be regarded as an auxiliary,
      non-differentiable output that will be ignored by the VJP function.

  Returns:
    A pair `(y, vjpfun)` if `has_aux` is False; a tuple `(y, vjpfun, aux)`
    otherwise. `y` and `aux` are the outputs of `f`, i.e. `y, aux =
    f(*primals)`. `vjpfun` is a function `dx = vjpfun(dy)`, where `dy` is the
    cotengents of `y`, having the same structures, shapes and dtypes as
    `y`. `dx` is the cotengents of `x`, having the same structures, shapes and
    dtypes as `x`.
  """
  with tf.GradientTape(persistent=True) as tape:
    tape.watch(tf.nest.flatten(primals))
    outputs = f(*primals)
    if has_aux:
      np_out, aux = outputs
    else:
      np_out = outputs

    def _vjp(dy):
      tf_dx = tape.gradient(np_out, primals, output_gradients=dy)
      return _tf_to_np(tf_dx)

  if has_aux:
    ret = (np_out, _vjp, aux)
  else:
    ret = (np_out, _vjp)
  return ret


# TODO(wangpeng): match JAX's handling of kwargs and non-ndarray args
def grad(f, has_aux=False):
  """Returns a function that computes gradient of f.

  Gradients can only be computed through numpy and tensorflow operations and not
  through python float operations and values.

  Args:
    f: a function of type (params, *args) -> scalar. 'params' can be a nested
      structure (made of lists and tuples) of ndarrays and the gradient is
      evaluated against it. `scalar` is a scalar ndarray.
    has_aux: bool, indicates whether fun returns a pair where the first element
      is considered the output of the mathematical function to be differentiated
      and the second element is auxiliary data.

  Returns:
    A gradient function of type (params, *args) -> gradients, where the result
    'gradients' has the same structure and shapes as 'params'.
  """

  def check_loss_shape(np_loss):
    if not isinstance(np_loss, tf_np.ndarray):
      raise ValueError(
          "The result of the function to take gradient must be an ndarray.")
    if not np_loss.shape.is_compatible_with([]):
      raise ValueError(
          "The result of the function to take gradient must be a scalar.")

  def _f(params, *args):
    """The gradient function to be returned."""
    with tf.GradientTape() as g:
      g.watch(tf.nest.flatten(params))
      outputs = f(params, *args)
      if has_aux:
        np_loss, aux = outputs
      else:
        np_loss = outputs
      check_loss_shape(np_loss)
      tf_grads = g.gradient(np_loss, params)
      if has_aux:
        res = (tf_grads, aux)
      else:
        res = tf_grads
      return _tf_to_np(res)

  return _f


def _record_result_type(recorder, f):
  """A decorator that records some information about the function.

  Args:
    recorder: a function of signature `(args, kwargs, res) -> res`.
    f: the original function.

  Returns:
    A transformed function that calls the original function and then the
    recorder afterwards.
  """
  def wrapper(*args, **kwargs):
    res = f(*args, **kwargs)
    res = recorder(args, kwargs, res)
    return res

  return wrapper


def jit(f,
        static_argnums=(),
        xla_forced_compile=False,
        input_signature=None,
        autograph=False,
        experimental_compile=False):
  """Returns a function that runs a trace-compiled version of `f`.

  A trace-compiled version of a function `f` has the same behavior as `f` (when
  called with the same "static arguments", see below), but runs faster because
  the whole computation is compiled into a computation graph once which is
  reused for subsequent executions.

  The trace compilation happens lazily, when the returned function is called for
  the first time. The compiled function may not be cached implicitly and
  multiple calls to `jit` may not share the compiled function (see below for
  "static" vs "dynamic" arguments).

  Args:
    f: a function that takes any positional arguments `args` and any keyword
      arguments `kwargs`. `ndarray`s and things accepted by
      `tf.convert_to_tensor` in `args` and `kwargs` will be treated as 'dynamic
      arguments' in the sense that calling the function with different values
      for these arguments will not cause retracing. In contrast, arguments of
      other types in `args` and `kwargs` are treated as 'static arguments' and
      calling the function with different values of them will cause
      re-compiling. Positional arguments whose positions are in `static_argnums`
      are always treated as static arguments.
    static_argnums: a tuple of positions of arguments that will be treated as
      static arguments. Note that as aforementioned, any arguments that were not
      convertible to tensor will also be static.
    xla_forced_compile: if true, it will use XLA to force-compile the graph.
      This requires that the function only contain ops that are XLA
      compatible. It will compile the entire function into a single XLA op.
    input_signature: a list of `tf.TensorSpec`, as the input signature to
      control tracing behavior. See the
      [doc](https://www.tensorflow.org/api_docs/python/tf/function]) of
        `tf.function` for details.
    autograph: whether to use autograph to convert Python constructs such as
      `if` and `while` to their TensorFlow counterparts. See the
      [doc](https://www.tensorflow.org/api_docs/python/tf/function]) of
        `tf.function` for details.
    experimental_compile: the `experimental_compile` flag for `tf.function`. See
      the [doc](https://www.tensorflow.org/api_docs/python/tf/function]) of
      `tf.function` for details. This is the recommended way to turn on XLA for
      tf.function, but unlike xla_forced_compile, it doesn't force-compile the
      entire function into a single XLA op.

  Returns:
    A trace-compiled version of f.
  """

  @tf.function(input_signature=input_signature, autograph=autograph,
               experimental_compile=experimental_compile)
  def _tf_f(*args, **kwargs):
    """Accelerated function with tensor inputs/outputs."""
    np_args = _tf_to_np(args)
    kwargs = {k: _tf_to_np(v) for k, v in kwargs.items()}
    if xla_forced_compile:
      # Use list for mutability
      output_is_list = [False]
      output_is_empty = [False]
      output_structure = [None]
      def recorder(args, kwargs, res):
        del args, kwargs
        # Workaround b/121383831
        output_is_list[0] = isinstance(res, list)
        # If outputs are empty, xla.compile returns an `Operation`, which we
        # don't want.
        if tf.nest.flatten(res):
          output_is_empty[0] = False
          output_structure[0] = None
        else:
          output_is_empty[0] = True
          # Without deepcopy, xla.compile will change output_structure[0] to a
          # list of `Operation`.
          output_structure[0] = copy.deepcopy(res)
        return res
      f_ = _record_result_type(recorder, f)
      np_out = tf.xla.experimental.compile(lambda: f_(*np_args, **kwargs))
      # Workaround b/121383831
      if output_is_empty[0]:
        np_out = output_structure[0]
      elif (isinstance(np_out, list) and len(np_out) == 1 and
            not output_is_list[0]):
        np_out = np_out[0]
    else:
      np_out = f(*np_args, **kwargs)
    return np_out

  def _f(*args, **kwargs):
    args = [
        _canonicalize_jit_arguments(arg) if i not in static_argnums else arg
        for i, arg in enumerate(args)
    ]
    kwargs = {k: _canonicalize_jit_arguments(v) for k, v in kwargs.items()}
    tf_out = _tf_f(*args, **kwargs)
    return _tf_to_np(tf_out)

  _f.tf_function = _tf_f

  return _f


def eval_on_shapes(f, static_argnums=(), allow_static_outputs=False):
  """Returns a function that evaluates `f` given input shapes and dtypes.

  It transforms function `f` to a function that performs the same computation as
  `f` but only on shapes and dtypes (a.k.a. shape inference).

  Args:
    f: the function to be transformed.
    static_argnums: see documentation of `jit`.
    allow_static_outputs: whether to allow non-array outputs. If True, non-array
      outputs (e.g. Python integers) will be returned as-is; otherwise, they
      will be converted to ndarrays, and then specs of those ndarrays will be
      returned.

  Returns:
    A function whose input arguments can be either the same as `f`'s or only
    their shapes/dtypes represented by `tf.TensorSpec`, and whose return values
    are `tf.TensorSpec`s with the same nested structure as `f`'s return
    values. If `allow_static_outputs` is True, when `f` returns some non-array
    outputs (e.g. Python integers), the converted function will return them
    as-is instead of returning `tf.TensorSpec`s for them.
  """
  def abstractify(args):
    def _abstractify(x):
      x = _canonicalize_jit_arg(x)
      if isinstance(x, (tf.Tensor, tf_np.ndarray)):
        return tf.TensorSpec(x.shape, x.dtype)
      else:
        return x
    new_args = []
    for i, arg in enumerate(args):
      if i in static_argnums:
        new_args.append(arg)
      else:
        new_args.append(tf.nest.map_structure(_abstractify, arg))
    return new_args

  if allow_static_outputs:
    # When `tf_f` below is called (via get_concrete_function) with the same
    # arugments (after abstraction), the Python function `f` won't be run, so we
    # need this python_outputs_map to retrieve the Python outputs we've seen
    # before that correspond the arguments.
    python_outputs_map = {}
    def recorder(args, kwargs, res):
      # Since the get_concrete_function below only uses positional args, we also
      # only positional args here.
      del args, kwargs
      def is_tensor_like(x):
        if hasattr(x, "_type_spec"):
          return True  # x is a CompositeTensor
        return isinstance(x, (tf_np.ndarray, tf.Tensor))
      py_values = tf.nest.map_structure(
          lambda x: None if is_tensor_like(x) else x,
          res)
      key = id(tf.compat.v1.get_default_graph())
      python_outputs_map[key] = py_values
      # Set non-tensor outputs to None to avoid tf.function calling
      # tf.convert_to_tensor on them.
      res = tf.nest.map_structure(
          lambda x: None if not is_tensor_like(x) else x,
          res)
      return res
    f = _record_result_type(recorder, f)

  # TODO(wangpeng): tf.function could add a knob to turn off materializing the
  #   graph, so that we don't waste computation and memory when we just want
  #   shape inference.
  tf_f = jit(f, static_argnums=static_argnums).tf_function

  # pylint: disable=missing-docstring
  def f_return(*args):
    def to_tensor_spec(x):
      if isinstance(x, tf.Tensor):
        return tf.TensorSpec(x.shape, x.dtype)
      else:
        return x

    new_args = abstractify(args)
    cfun = tf_f.get_concrete_function(*new_args)
    res = cfun.structured_outputs
    res = tf.nest.map_structure(to_tensor_spec, res)

    if allow_static_outputs:
      key = id(cfun.graph)
      py_values = python_outputs_map[key]
      # We can also call tf.get_static_value on structured_outputs to retrieve
      # the Python values, but since we'll need to use python_outputs_map to
      # record "which outputs are static?" anyway, we choose to directly store
      # the Python values in python_outputs_map.
      res = tf.nest.map_structure(
          lambda x, python_value: x if python_value is None else python_value,
          res, py_values)

    return res

  # Provides access to `tf_f` for testing purpose.
  f_return._tf_function = tf_f  # pylint: disable=protected-access
  return f_return


def _index_update_helper(updater, x, idx, y):
  x = tf_np.asarray(x)
  y = tf_np.asarray(y)
  # TODO(b/164251540): Remove this expensive manual broadcasting once
  #   tf.raw_ops.tensor_strided_slice_update and tf.tensor_scatter_nd_update
  #   support broadcasting.
  y = tf.broadcast_to(y, tf.shape(x[idx]))
  return updater(x, idx, y)


# pylint: disable=protected-access
def index_update(x, idx, y):
  """Pure equivalent of `x[idx] = y`.

  Returns the value of x that would result from the NumPy-style indexed
  assignment `x[idx] = y`. Because it's a pure function, `x` itself won't be
  changed.

  Args:
    x: an array with the values to be updated.
    idx: a Numpy-style index, consisting of `None`, integers, slice objects,
      ellipses, ndarrays with integer dtypes, or a tuple of the above.
    y: the array of updates. `y` must be broadcastable to the shape of the array
      that would be returned by `x[idx]`.

  Returns:
    The updated version of `x`.
  """
  return _index_update_helper(tf_np.ndarray._with_index_update, x, idx, y)


def index_add(x, idx, y):
  """Pure equivalent of `x[idx] += y`.

  Returns the value of x that would result from the NumPy-style indexed
  assignment `x[idx] += y`. Because it's a pure function, `x` itself won't be
  changed.

  Args:
    x: an array with the values to be updated.
    idx: a Numpy-style index, consisting of `None`, integers, slice objects,
      ellipses, ndarrays with integer dtypes, or a tuple of the above.
    y: the array of updates. `y` must be broadcastable to the shape of the array
      that would be returned by `x[idx]`.

  Returns:
    The updated version of `x`.
  """
  return _index_update_helper(tf_np.ndarray._with_index_add, x, idx, y)


def index_min(x, idx, y):
  """Pure equivalent of `x[idx] = minimum(x[idx], y)`.

  Returns the value of x that would result from the NumPy-style indexed
  assignment `x[idx] = minimum(x[idx], y)`. Because it's a pure function, `x`
  itself won't be changed.

  Args:
    x: an array with the values to be updated.
    idx: a Numpy-style index, consisting of `None`, integers, slice objects,
      ellipses, ndarrays with integer dtypes, or a tuple of the above.
    y: the array of updates. `y` must be broadcastable to the shape of the array
      that would be returned by `x[idx]`.

  Returns:
    The updated version of `x`.
  """
  return _index_update_helper(tf_np.ndarray._with_index_min, x, idx, y)


def index_max(x, idx, y):
  """Pure equivalent of `x[idx] = maximum(x[idx], y)`.

  Returns the value of x that would result from the NumPy-style indexed
  assignment `x[idx] = maximum(x[idx], y)`. Because it's a pure function, `x`
  itself won't be changed.

  Args:
    x: an array with the values to be updated.
    idx: a Numpy-style index, consisting of `None`, integers, slice objects,
      ellipses, ndarrays with integer dtypes, or a tuple of the above.
    y: the array of updates. `y` must be broadcastable to the shape of the array
      that would be returned by `x[idx]`.

  Returns:
    The updated version of `x`.
  """
  return _index_update_helper(tf_np.ndarray._with_index_max, x, idx, y)
# pylint: enable=protected-access


def logsumexp(x, axis=None, keepdims=None):
  """Computes log(sum(exp(elements across dimensions of a tensor))).

  Reduces `x` along the dimensions given in `axis`.
  Unless `keepdims` is true, the rank of the tensor is reduced by 1 for each
  entry in `axis`. If `keepdims` is true, the reduced dimensions
  are retained with length 1.
  If `axis` has no entries, all dimensions are reduced, and a
  tensor with a single element is returned.
  This function is more numerically stable than log(sum(exp(input))). It avoids
  overflows caused by taking the exp of large inputs and underflows caused by
  taking the log of small inputs.

  Args:
    x: The tensor to reduce. Should have numeric type.
    axis: The dimensions to reduce. If `None` (the default), reduces all
      dimensions. Must be in the range `[-rank(x), rank(x))`.
    keepdims: If true, retains reduced dimensions with length 1.

  Returns:
    The reduced tensor.
  """
  return tf_np.asarray(
      tf.math.reduce_logsumexp(
          input_tensor=x, axis=axis, keepdims=keepdims))


def expit(x):
  """Compute 1 / (1 + exp(-x))."""
  return tf_np.asarray(tf.math.sigmoid(x))


def erf(x):
  """Computes the Gauss error function of x element-wise."""
  return tf_np.asarray(tf.math.erf(x))


def _minus(a, b):
  return [x for x in a if x not in b]


def _compose_output_rep(lhs_rep, rhs_rep, lhs_contraction, rhs_contraction,
                        lhs_batch, rhs_batch):
  """Compose the output string representation.

  e.g., ij, jk, (((1,), (0,)), ((), ())) -> ik
        aij, ajk, (((2,), (1,)), ((0,), (0,))) -> aik

  Args:
    lhs_rep: A string representation for the left-hand side input array
    rhs_rep: A string representation for the right-hand side input array
    lhs_contraction: Sequence[int] (the contraction dimensions of lhs)
    rhs_contraction: Sequence[int] (the contraction dimensions of rhs)
    lhs_batch: Sequence[int] (the batch dimensions of lhs)
    rhs_batch: Sequence[int] (the batch dimensions of rhs)

  Returns:
    A string representation of the result array.
  """
  output_rep = []
  for dim in lhs_batch:
    output_rep.append(lhs_rep[dim])

  for i in _minus(range(len(lhs_rep)), lhs_batch + lhs_contraction):
    output_rep.append(lhs_rep[i])
  for i in _minus(range(len(rhs_rep)), rhs_batch + rhs_contraction):
    output_rep.append(rhs_rep[i])
  return "".join(output_rep)


def _non_batched_matmul(lhs, rhs, lhs_contraction, rhs_contraction):
  """Compute the non-batched matrix multiplication.

  If it is the general non-batched/single-batched matrix multiplication,
  use the highly optimized kernel `tf.tensordot` to handle it.

  Args:
    lhs: an array (the left-hand side matrix/vector to be multiplied)
    rhs: an array (the right-hand side matrix/vector to be multiplied)
    lhs_contraction: Sequence[int] (the contraction dimensions of lhs)
    rhs_contraction: Sequence[int] (the contraction dimensions of rhs)

  Returns:
    An array that contains the result.
  """
  return tf.tensordot(
      lhs, rhs, axes=(list(lhs_contraction), list(rhs_contraction)))


def tf_dot_general(lhs, rhs, dimension_numbers):
  """The general dot operation for TensorFlow.

  An equivalent general dot operation as that in JAX -
     <https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.dot_general.html>
  Although there is an implementation in TF XLA, avoid directly using XLA when
  possible.

  e.g., non-batched: ij,jk->ik
        batched: ijk,ikl->ijl

  Args:
    lhs: an array (the left-hand side matrix/vector to be multiplied)
    rhs: an array (the right-hand side matrix/vector to be multiplied)
    dimension_numbers: (Tuple[Tuple[Sequence[int], Sequence[int]],
      Tuple[Sequence[int], Sequence[int]]]) – a tuple of tuples of the form
      ((lhs_contracting_dims, rhs_contracting_dims), (lhs_batch_dims,
      rhs_batch_dims))

  Returns:
    An array that contains the result.
  """
  char_list = list(string.ascii_lowercase)
  char_list = char_list[8:] + char_list[:8]
  lhs_rank, rhs_rank = len(lhs.shape), len(rhs.shape)
  lhs_rep = char_list[:lhs_rank]
  rhs_rep = char_list[lhs_rank:lhs_rank + rhs_rank]
  contraction, batch = dimension_numbers
  lhs_contraction, rhs_contraction = contraction
  if len(lhs_contraction) != len(rhs_contraction):
    raise ValueError(
        "The input matrices are required to have the same number "
        "of contraction dimensions, but got: lhs {}, rhs: {}".format(
            len(lhs_contraction), len(rhs_contraction)))
  lhs_batch, rhs_batch = batch
  if len(lhs_batch) != len(rhs_batch):
    raise ValueError("The input matrices are required to have the same number "
                     "of batch dimensions, but got: lhs {}, rhs: {}".format(
                         len(lhs_batch), len(rhs_batch)))

  if not lhs_batch and not rhs_batch:
    return _non_batched_matmul(lhs, rhs, lhs_contraction, rhs_contraction)

  if (lhs_rank == rhs_rank == 3 and lhs_batch == (0,) and rhs_batch == (0,) and
      lhs_contraction == (2,) and rhs_contraction == (1,)):
    return tf.linalg.matmul(lhs, rhs)

  for i in range(len(lhs_contraction)):
    rhs_rep[rhs_contraction[i]] = lhs_rep[lhs_contraction[i]]
  for i in range(len(lhs_batch)):
    rhs_rep[rhs_batch[i]] = lhs_rep[lhs_batch[i]]

  output_rep = _compose_output_rep(lhs_rep, rhs_rep, lhs_contraction,
                                   rhs_contraction, lhs_batch, rhs_batch)
  equation = "".join(lhs_rep) + "," + "".join(rhs_rep) + "->" + output_rep
  return tf.einsum(equation, lhs, rhs)


def _conv_general_param_type_converter(window_strides, lhs_dilation,
                                       rhs_dilation, dim):
  """Convert strides, lhs_dilation, rhs_dilation to match TF convention.

  For example,
   in the 3D case, if lhs_dilation = 2, then convert it to [2, 2, 2]
                   if lhs_dilation = (2, 2, 2), convert it also to [2, 2, 2]

  Args:
    window_strides: window_strides to be converted
    lhs_dilation: lhs_dilation to be converted
    rhs_dilation: rhs_dilation to be converted
    dim: dim to be converted

  Returns:
    The updated window_strides, lhs_dilation and rhs_dilation
  """
  def _as_list_of_size(item, size):
    if item is None:
      return None
    return [item] * size if isinstance(item, int) else list(item)
  return (_as_list_of_size(window_strides, dim),
          _as_list_of_size(lhs_dilation, dim),
          _as_list_of_size(rhs_dilation, dim))


# pylint: disable=g-bad-todo
# TODO(DarrenZhang01): Expand the test cases of general convolution and revise
# the according bugs.
# TODO(DarrenZhang01): Support feature_group_count, batch_group_count and
# precision, and allow lhs_dilation and rhs_dilation to happen at the same time.
# pylint: enable=g-bad-todo
def tf_conv_general_dilated(lhs, rhs, window_strides, padding, output_shape,
                            lhs_dilation=None, rhs_dilation=None,
                            dimension_numbers=None, feature_group_count=1,
                            batch_group_count=1, precision=None):
  """A general conv API for TensorFlow.

  According JAX version:
    https://jax.readthedocs.io/en/stable/_autosummary/jax.lax.conv_general_dilated.html

  Args:
    lhs: a rank n+2 dimensional input array.
    rhs: a rank n+2 dimensional array of kernel weights.
    window_strides: a sequence of n integers, representing the inter-window
                    strides.
    padding: either the string ‘SAME’, the string ‘VALID’, or a sequence of n
             (low, high) integer pairs that give the padding to apply before and
             after each spatial dimension.
    output_shape: the output shape of the convolution (only required for
                  transpose convolution).
    lhs_dilation: None, or a sequence of n integers, giving the dilation factor
                  to apply in each spatial dimension of lhs. LHS dilation is
                  also known as transposed convolution.
    rhs_dilation: None, or a sequence of n integers, giving the dilation factor
                  to apply in each spatial dimension of rhs. RHS dilation is
                  also known as atrous convolution.
    dimension_numbers: either None, a ConvDimensionNumbers object, or a 3-tuple
                       (lhs_spec, rhs_spec, out_spec), where each element is a
                       string of length n+2.
    feature_group_count:  integer, default 1. Changing this is currently not
                          supported.
    batch_group_count: integer, default 1. Changing this is currently not
                       supported.
    precision: Optional. Either None, which means the default precision for the
               backend, or a Precision enum value.

  Returns:
    A TF NumPy array that contains the convolution result.
  """
  dim = None
  lhs_spec, rhs_spec, out_spec = dimension_numbers
  if lhs_spec != out_spec:
    raise ValueError("Current implementation requires the `data_format` of the "
                     "inputs and outputs to be the same.")
  if len(lhs_spec) >= 6:
    raise ValueError("Current implmentation does not support 4 or higher"
                     "dimensional convolution, but got: ", len(lhs_spec) - 2)
  dim = len(lhs_spec) - 2
  if lhs_dilation and rhs_dilation:
    if lhs_dilation == (1,) * dim and rhs_dilation == (1,) * dim:
      lhs_dilation, rhs_dilation = None, None
    else:
      raise ValueError("Current implementation does not support that "
                       "deconvolution and dilation to be performed at the same "
                       "time, but got lhs_dilation: {}, rhs_dilation: {}"
                       .format(lhs_dilation, rhs_dilation))
  if padding not in ["SAME", "VALID"]:
    raise ValueError("Current implementation requires the padding parameter"
                     "to be either 'VALID' or 'SAME', but got: ", padding)
  if batch_group_count != 1 or feature_group_count != 1:
    raise NotImplementedError("batch_group_count and feature_group_count "
                              "other than 1 is currently not supported, but"
                              " got feature_group_count: {}, batch_group_count"
                              ": {}".format(feature_group_count,
                                            batch_group_count))
  if precision is not None:
    raise NotImplementedError("precision other than `None` is currently not "
                              "supported, but got: {}".format(precision))
  # Convert params from int/Sequence[int] to list of ints.
  strides, lhs_dilation, rhs_dilation = _conv_general_param_type_converter(
      window_strides, lhs_dilation, rhs_dilation, dim
  )
  # Preprocess the shapes
  dim_maps = {}
  if isinstance(lhs_spec, str):
    dim_maps["I"] = list(rhs_spec).index("I")
    dim_maps["O"] = list(rhs_spec).index("O")
    dim_maps["N"] = list(lhs_spec).index("N")
    dim_maps["C"] = list(lhs_spec).index("C")
  else:
    dim_maps["I"] = rhs_spec[1]
    dim_maps["O"] = rhs_spec[0]
    dim_maps["N"] = lhs_spec[0]
    dim_maps["C"] = lhs_spec[1]

  lhs = tf_np.moveaxis(lhs, (dim_maps["N"], dim_maps["C"]), (0, dim + 1))
  # Adjust the filters, put the dimension 'I' and 'O' at last.
  rhs = tf_np.moveaxis(rhs, (dim_maps["O"], dim_maps["I"]), (dim + 1, dim))
  spatial_dim_maps = {1: "W", 2: "HW", 3: "DHW"}
  data_format = "N" + spatial_dim_maps[dim] + "C"

  if rhs_dilation or (lhs_dilation is None and rhs_dilation is None):
    output = _tf_nn_APIs[dim][0](lhs, rhs, strides, padding, data_format,
                                 rhs_dilation)
  else:
    output = _tf_nn_APIs[dim][1](lhs, rhs, tf.constant(output_shape), strides,
                                 padding, data_format, lhs_dilation)
  output = tf_np.moveaxis(output, (0, dim + 1), (dim_maps["N"], dim_maps["C"]))
  return output


def conv(inp,
         fltr,
         window_strides,
         padding,
         dimension_numbers,
         filter_dilation=None):
  """Convolution over an N-D array.

  See https://www.tensorflow.org/api_docs/python/tf/nn/convolution and
  https://www.tensorflow.org/xla/operation_semantics#conv_convolution for
  reference.

  Args:
    inp: an (N+2)-D array. The input of the convolution.
    fltr: an (N+2)-D array. The filter (i.e. kernel) of the convolution.
    window_strides: a sequence of N ints, the strides for moving the convolution
      window.
    padding: a string, either "VALID" or "SAME". The padding algorithm.
    dimension_numbers: a tuple of three strings encoding the data format of
      input, filter and output. "I" means input; "O" means output; "C" means
      channel; other characters such as "W", "H" and "D" means spatial
      dimensions.
    filter_dilation: the dilation rates for the filter. Dilating the filter
      means adding "holes" to the filter.

  Returns:
    An (N+2)-D array. The convolution result.
  """
  input_spec, filter_spec, output_spec = dimension_numbers
  if input_spec != output_spec:
    raise ValueError("Input and output data formats must be the same; got %s "
                     "and %s" % (input_spec, output_spec))
  supported_filter_spec = ["WIO", "HWIO", "DHWIO"]
  if filter_spec not in supported_filter_spec:
    raise ValueError("The supported data format for the filter are %s; got %s" %
                     (supported_filter_spec, filter_spec))
  if input_spec[1:-1] != filter_spec[:-2]:
    raise ValueError("Input data format (%s) is not compatible with filter "
                     "data format (%s)" % (input_spec, filter_spec))
  # No type promotion in order to prevent accidentally doing more expensive
  # computation.
  dtype = tf_np.result_type(inp, fltr)
  inp = tf_np.asarray(inp, dtype)
  fltr = tf_np.asarray(fltr, dtype)
  return tf_np.asarray(
      tf.nn.convolution(
          input=inp,
          filters=fltr,
          padding=padding,
          strides=window_strides,
          dilations=filter_dilation,
          data_format=input_spec))


def avg_pool(x, pool_size, strides, padding):
  """Performs an N-D average pooling.

  Args:
    x: ndarray of rank N+2, of shape `[batch_size] + input_spatial_shape +
      [num_channels]`. Pooling happens over the spatial dimensions only.
    pool_size: sequence of N ints.
    strides: sequence of N ints.
    padding: a string, the padding algorithm. Must be "SAME" or "VALID".

  Returns:
    An (N+2)-D array,  of shape
      [batch_size] + output_spatial_shape + [num_channels],
    where `output_spatial_shape` depends on the value of padding:
    If padding = "SAME":
      output_spatial_shape[i] = ceil(input_spatial_shape[i] / strides[i])
    If padding = "VALID":
      output_spatial_shape[i] =
        ceil((input_spatial_shape[i] - (pool_size[i] - 1)) / strides[i]).
  """
  x = tf_np.asarray(x)
  return tf_np.asarray(
      tf.nn.pool(
          input=x,
          window_shape=pool_size,
          pooling_type="AVG",
          strides=strides,
          padding=padding))


def max_pool(x, pool_size, strides, padding):
  """Performs an N-D max pooling.

  Args:
    x: ndarray of rank N+2, of shape `[batch_size] + input_spatial_shape +
      [num_channels]`. Pooling happens over the spatial dimensions only.
    pool_size: sequence of N ints.
    strides: sequence of N ints.
    padding: a string, the padding algorithm. Must be "SAME" or "VALID".

  Returns:
    An (N+2)-D array,  of shape
      [batch_size] + output_spatial_shape + [num_channels],
    where `output_spatial_shape` depends on the value of padding:
    If padding = "SAME":
      output_spatial_shape[i] = ceil(input_spatial_shape[i] / strides[i])
    If padding = "VALID":
      output_spatial_shape[i] =
        ceil((input_spatial_shape[i] - (pool_size[i] - 1)) / strides[i]).
  """
  x = tf_np.asarray(x)
  return tf_np.asarray(
      tf.nn.pool(
          input=x,
          window_shape=pool_size,
          pooling_type="MAX",
          strides=strides,
          padding=padding))


def sort_key_val(keys, values, dimension=-1):
  """Sorts keys along a dimension and applies same permutation to values.

  Args:
    keys: an array. The dtype must be comparable numbers (integers and reals).
    values: an array, with the same shape of `keys`.
    dimension: an `int`. The dimension along which to sort.

  Returns:
    Permuted keys and values.
  """
  keys = tf_np.asarray(keys)
  values = tf_np.asarray(values)
  rank = keys.shape.ndims
  if rank is None:
    rank = values.shape.ndims
  if rank is None:
    # We need to know the rank because tf.gather requires batch_dims to be `int`
    raise ValueError("The rank of either keys or values must be known, but "
                     "both are unknown (i.e. their shapes are both None).")
  if dimension in (-1, rank - 1):

    def maybe_swapaxes(a):
      return a
  else:

    def maybe_swapaxes(a):
      return tf_np.swapaxes(a, dimension, -1)

  # We need to swap axes because tf.gather (and tf.gather_nd) supports
  # batch_dims on the left but not on the right.
  # TODO(wangpeng): Investigate whether we should do swapaxes or moveaxis.
  keys = maybe_swapaxes(keys)
  values = maybe_swapaxes(values)
  idxs = tf_np.argsort(keys)

  # Using tf.gather rather than np.take because the former supports batch_dims
  def gather(a):
    return tf_np.asarray(tf.gather(a, idxs, batch_dims=rank - 1))

  keys = gather(keys)
  values = gather(values)
  keys = maybe_swapaxes(keys)
  values = maybe_swapaxes(values)
  return keys, values


def scan(f, init, xs, length=None, reverse=False):
  """Scan a function over leading array axes while carrying along state.

  See the docstring of `jax.lax.scan`
  (https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.scan.html) for
  details.

  Args:
    f: a Python function to be scanned of type ``c -> a -> (c, b)``, meaning
      that ``f`` accepts two arguments where the first is a value of the loop
      carry and the second is a slice of ``xs`` along its leading axis, and that
      ``f`` returns a pair where the first element represents a new value for
      the loop carry and the second represents a slice of the output. Note that
      the input and output carry must have the same dtype.
    init: an initial loop carry value of type ``c``, which can be a scalar,
      array, or any pytree (nested Python tuple/list/dict) thereof, representing
      the initial loop carry value. This value must have the same structure as
      the first element of the pair returned by ``f``.
    xs: the value of type ``[a]`` over which to scan along the leading axis,
      where ``[a]`` can be an array or any pytree (nested Python
      tuple/list/dict) thereof with consistent leading axis sizes.
    length: optional integer specifying the number of loop iterations, which
      must agree with the sizes of leading axes of the arrays in ``xs`` (but can
      be used to perform scans where no input ``xs`` are needed).
    reverse: optional boolean specifying whether to run the scan iteration
      forward (the default) or in reverse, equivalent to reversing the leading
      axes of the arrays in both ``xs`` and in ``ys``.

  Returns:
    A pair of type ``(c, [b])`` where the first element represents the final
    loop carry value and the second element represents the stacked outputs of
    the second output of ``f`` when scanned over the leading axis of the inputs.
  """
  init, xs = tf.nest.map_structure(
      lambda x: tf_np.asarray(x) if x is not None else None, (init, xs))
  if length is not None:
    length = int(length)
  def get_length(x):
    if x is None:
      return None
    if x.shape.rank == 0:
      raise ValueError("Some array in `xs` doesn't have a leading dimension")
    return x.shape[0]
  lengths = tf.nest.flatten(tf.nest.map_structure(get_length, xs))
  for l in lengths:
    if l is not None:
      if length is None:
        length = l
      elif length != l:
        raise ValueError("There are two different leading-dimension lengths: "
                         f"{length} and {l}")
  if length is None:
    raise ValueError(
        "Can't determine length. Please set the `length` argument.")
  xs_ta = tf.nest.map_structure(
      lambda t: (tf.TensorArray(t.dtype, size=length, dynamic_size=False)  # pylint: disable=g-long-lambda
                 .unstack(t) if t is not None else None),
      xs)
  # tf.while_loop doesn't allow None in loop_vars, so we mask them.
  is_init_none = tf.nest.map_structure(lambda x: x is None, init)
  def to_safe(carry):
    return tf.nest.map_structure(
        lambda x, is_none: tf.zeros([]) if is_none else x, carry, is_init_none)
  def from_safe(safe_carry):
    return tf.nest.map_structure(
        lambda x, is_none: None if is_none else x, safe_carry, is_init_none)
  def body(i, safe_carry, ys_ta):
    carry = from_safe(safe_carry)
    if reverse:
      i_ = length - 1 - i
    else:
      i_ = i
    xs = tf.nest.map_structure(
        lambda x_ta: x_ta.read(i_) if x_ta is not None else None, xs_ta)
    carry, ys = f(*_tf_to_np((carry, xs)))
    ys_ta = tf.nest.map_structure(
        lambda y_ta, y: (y_ta.write(i_, y) if y is not None else y_ta),
        ys_ta, ys)
    i = i + 1
    safe_carry = to_safe(carry)
    return i, safe_carry, ys_ta
  xs_spec = tf.nest.map_structure(
      lambda t: tf.TensorSpec(t.shape[1:], t.dtype) if t is not None else None,
      xs)
  _, ys_spec = eval_on_shapes(f)(init, xs_spec)
  # ys_ta can't contain None because tf.while_loop doesn't allow None in
  # loop_vars.
  ys_ta = tf.nest.map_structure(
      lambda y: tf.TensorArray(y.dtype if y is not None else tf.float32,  # pylint: disable=g-long-lambda
                               size=length, dynamic_size=False),
      ys_spec)
  safe_init = to_safe(init)
  _, safe_carry, ys_ta = tf.while_loop(
      lambda i, *_: i < length, body, (0, safe_init, ys_ta),
      maximum_iterations=length)
  carry = from_safe(safe_carry)
  def _stack(a, spec):
    if spec is None:
      return None
    a = a.stack()
    a.set_shape((length,) + a.shape[1:])
    return a
  ys = tf.nest.map_structure(_stack, ys_ta, ys_spec)
  return _tf_to_np((carry, ys))


# named "tf_map" instead of "map" as in JAX to avoid conflict with Python `map`
def tf_map(f, xs):
  """Map a function over leading array axes.

  See the docstring of `jax.lax.map`
  (https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.map.html) for
  details.

  Args:
    f: a Python function to apply element-wise over the first axis or axes of
      `xs`.
    xs: values over which to map along the leading axis.

  Returns:
    Mapped values.
  """
  def g(unused, x):
    return unused, f(x)
  carry = tf.nest.map_structure(lambda _: None, xs)
  return scan(g, carry, xs)[1]


def _get_dynamic_indices(operand, start_indices, slice_sizes):
  """Calcuates the indices for `tf.gather_nd` from slices.

  Args:
    operand: a Tensor to slice.
    start_indices: a vector Tensor of integers, one per dimension. The starts of
      the slice. The vector can be dynamic.
    slice_sizes: a list of integers, one per dimension. The sizes of the slice.

  Returns:
    An index array suitable for `tf.gather_nd` and `tf.scatter_nd`, or `None` if
    `operand` is a scalar.
  """
  rank = len(slice_sizes)
  operand_rank = tf.rank(operand)
  tf.debugging.Assert(operand_rank == rank, [operand_rank, rank])
  starts_rank = tf.rank(start_indices)
  tf.debugging.Assert(starts_rank == 1, [starts_rank])
  num_starts = tf.shape(start_indices)[0]
  tf.debugging.Assert(num_starts == rank, [num_starts, rank])
  operand_shape = tf.shape(operand)
  tf.debugging.Assert(tf.reduce_all(slice_sizes <= operand_shape),
                      [slice_sizes, operand_shape])
  if rank == 0:
    return None
  start_indices = tf.where(
      start_indices < 0, start_indices + operand_shape, start_indices)
  idx_list = []
  for i in range(rank):
    start = start_indices[i]
    size = slice_sizes[i]
    dim = operand_shape[i]
    start = tf.clip_by_value(start, 0, dim - size)
    # XLA requires tf.range's `start` to be compile-time constant, so we can't
    # do tf.range(start, ...).
    idx = start + tf.range(size)
    shape = [1] * rank
    shape[i] = size
    idx = tf.reshape(idx, shape)
    idx_list.append(idx)
  slice_sizes_tensor = tf.convert_to_tensor(slice_sizes)
  # tf.stack doesn't support broadcasting, so we need to broadcast manually.
  # TODO(wangpeng): Reduce peak memory by broadcasting one-by-one instead of
  #   all-together.
  idx_list = [tf.broadcast_to(x, slice_sizes_tensor) for x in idx_list]
  return tf.stack(idx_list, axis=-1)


def dynamic_slice(operand, start_indices, slice_sizes):
  """Slicing operation where the indices can be dynamic vlaues.

  See the docstring of `jax.lax.dynamic_slice`
  (https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.dynamic_slice.html)
  for details.

  Args:
    operand: an array to slice.
    start_indices: a vector of integers, one per dimension. The starts of the
      slice. The vector can be dynamic.
    slice_sizes: a list of integers, one per dimension. The sizes of the slice.

  Returns:
    An array containing the slice, with shape equal to `slice_sizes`.
  """
  # This implementation uses tf.gather_nd to implement dynamic_slice, which is
  # memory inefficient because the size of `indices` given to gather_nd is
  # large.
  operand = tf_np.asarray(operand).data
  start_indices = tf_np.asarray(start_indices, np.int32).data
  idx = _get_dynamic_indices(operand, start_indices, slice_sizes)
  if idx is not None:
    operand = tf.gather_nd(operand, idx)
  return tf_np.asarray(operand)


def dynamic_update_slice(operand, update, start_indices):
  """Updates a dynamic slice.

  See the docstring of `jax.lax.dynamic_update_slice`
  (https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.dynamic_update_slice.html)
  for details.

  Args:
    operand: an array to slice.
    update: an array containing the new values to write onto `operand`.
    start_indices: a vector of integers, one per dimension. The starts of the
      slice. The vector can be dynamic.

  Returns:
    The updated version of `operand`.
  """
  operand = tf_np.asarray(operand).data
  update = tf_np.asarray(update).data
  start_indices = tf_np.asarray(start_indices, np.int32).data
  if not update.shape.is_fully_defined():
    raise ValueError("update's shape must be fully defined")
  slice_sizes = update.shape
  idx = _get_dynamic_indices(operand, start_indices, slice_sizes)
  if idx is None:
    # `np.zeros([])[()] = 1.0` will result in a scalar array of 1.0
    return tf_np.asarray(update)
  operand = tf.tensor_scatter_nd_update(operand, idx, update)
  return tf_np.asarray(operand)


def dynamic_slice_in_dim(operand, start_index, slice_size, axis=0):
  """Convenience wrapper around dynamic_slice applying to one dimension."""
  operand = tf_np.asarray(operand)
  start_indices = [0] * operand.ndim
  slice_sizes = list(operand.shape)
  axis = int(axis)
  start_indices[axis] = start_index
  slice_sizes[axis] = int(slice_size)
  return dynamic_slice(operand, start_indices, slice_sizes)


def dynamic_update_slice_in_dim(operand, update, start_index, axis):
  """Convenience wrapper around dynamic_update_slice for one dimension."""
  operand = tf_np.asarray(operand)
  axis = int(axis)
  start_indices = [0] * operand.ndim
  start_indices[axis] = start_index
  return dynamic_update_slice(operand, update, start_indices)


# Use int64 instead of int32 to avoid TF's "int32 problem"
_RNG_KEY_DTYPE = np.int64


def _key2seed(a):
  """Converts an RNG key to an RNG seed.

  Args:
    a: an RNG key, an ndarray of shape [] and dtype `np.int64`.

  Returns:
    an RNG seed, a tensor of shape [2] and dtype `tf.int32`.
  """

  def int64_to_int32s(a):
    """Converts an int64 tensor of shape [] to an int32 tensor of shape [2]."""
    a = tf.cast(a, tf.uint64)
    fst = tf.cast(a, tf.uint32)
    snd = tf.cast(
        tf.bitwise.right_shift(a, tf.constant(32, tf.uint64)), tf.uint32)
    a = [fst, snd]
    a = tf.nest.map_structure(lambda x: tf.cast(x, tf.int32), a)
    a = tf.stack(a)
    return a

  return int64_to_int32s(a)


def _seed2key(a):
  """Converts an RNG seed to an RNG key.

  Args:
    a: an RNG seed, a tensor of shape [2] and dtype `tf.int32`.

  Returns:
    an RNG key, an ndarray of shape [] and dtype `np.int64`.
  """

  def int32s_to_int64(a):
    """Converts an int32 tensor of shape [2] to an int64 tensor of shape []."""
    a = tf.bitwise.bitwise_or(
        tf.cast(a[0], tf.uint64),
        tf.bitwise.left_shift(
            tf.cast(a[1], tf.uint64), tf.constant(32, tf.uint64)))
    a = tf.cast(a, tf.int64)
    return a

  return tf_np.asarray(int32s_to_int64(a))


def prng(s):
  """Creates RNG state from seed.

  Args:
    s: the seed, an integer.

  Returns:
    An RNG state, as a scalar array of dtype `np.int64`.
  """
  # TODO(wangpeng): Become bitwise-identical to JAX when TF stateless RNGs get
  #   improved.
  return tf_np.asarray(s, dtype=_RNG_KEY_DTYPE)


def stateless_split(seed, num=2):
  """Splits an RNG seed into `num` new seeds by adding a leading axis.

  Example:

  >>> seed = [1, 2]
  >>> new_seeds = tf.random.experimental.stateless_split(seed, num=3)
  >>> print(new_seeds)
  tf.Tensor(
  [[1105988140 1738052849]
   [-335576002  370444179]
   [  10670227 -246211131]], shape=(3, 2), dtype=int32)
  >>> tf.random.stateless_normal(shape=[3], seed=new_seeds[0, :])
  <tf.Tensor: shape=(3,), dtype=float32, numpy=array([-0.59835213, -0.9578608 ,
  0.9002807 ], dtype=float32)>

  Args:
    seed: an RNG seed (a tensor with shape [2] and dtype `int32` or `int64`).
      (When using XLA, only `int32` is allowed.)
    num: optional, a positive integer or scalar tensor indicating the number of
      seeds to produce (default 2).

  Returns:
    A tensor with shape [num, 2] representing `num` new seeds. It will have the
    same dtype as `seed` (if `seed` doesn't have an explict dtype, the dtype
    will be determined by `tf.convert_to_tensor`).
  """
  seed = tf.convert_to_tensor(seed)
  return tf.random.stateless_uniform(
      shape=[num, 2], seed=seed, dtype=seed.dtype, minval=None, maxval=None)


def split(state, num):
  """Creates new independent RNG states from an existing state.

  Args:
    state: the existing state.
    num: the number of the new states.

  Returns:
    A tuple of new states.
  """
  state = tf_np.asarray(state, dtype=_RNG_KEY_DTYPE)
  state = _key2seed(state)
  try:
    states = tf.random.experimental.stateless_split(state, num)
  except AttributeError as e:  # pylint: disable=unused-variable
    # TODO(afrozm): For TF < 2.3 we need to do this. Delete once 2.3 launches.
    states = stateless_split(state, num)
  states = tf.unstack(states, num)
  states = tf.nest.map_structure(_seed2key, states)
  return states


def uniform(key,
            shape,
            dtype=tf_np.random.DEFAULT_RANDN_DTYPE,
            minval=0.,
            maxval=1.):
  """Sample uniform random values in range [`minval`, `maxval`).

  Args:
    key: the RNG key.
    shape: the shape of the result.
    dtype: the dtype of the result.
    minval: the minimal value (inclusive).
    maxval: the maximal value (exclusive).

  Returns:
    An ndarray with shape `shape` and dtype `dtype`. Each value in the ndarray
    is sampled uniformly randomly in range [`minval`, `maxval`).
  """
  minval = tf.cast(minval, dtype)
  maxval = tf.cast(maxval, dtype)
  key = tf_np.asarray(key, dtype=_RNG_KEY_DTYPE)
  return tf_np.asarray(
      tf.random.stateless_uniform(
          shape, seed=_key2seed(key), dtype=dtype, minval=minval,
          maxval=maxval))


def normal(key, shape, dtype=tf.float32):
  """Sample standard-normal random values.

  Args:
    key: the RNG key.
    shape: the shape of the result.
    dtype: the dtype of the result.

  Returns:
    Random values in standard-normal distribution.
  """
  key = tf_np.asarray(key, dtype=_RNG_KEY_DTYPE)
  return tf_np.asarray(
      tf.random.stateless_normal(shape, seed=_key2seed(key), dtype=dtype))


def bernoulli(key, mean=np.float32(0.5), shape=None):
  """Sample Bernoulli random values with given shape and mean.

  Args:
    key: the RNG key.
    mean: optional, an array_like broadcastable to `shape` for the mean of the
      random variables (default 0.5).
    shape: optional, a tuple of nonnegative integers representing the shape
      (default to `mean`'s shape).

  Returns:
    A random array with the specified shape and boolean dtype.
  """
  mean = tf_np.asarray(mean)
  if shape is None:
    shape = mean.shape
  return uniform(key, shape) < mean


def _eager_dataset_iterator(dataset):
  for item in dataset:
    yield tf.nest.map_structure(tf_np.asarray, item)


def dataset_as_numpy(dataset):
  """Converts a `tf.data.Dataset` to an iterable of ndarrays.

  `dataset_as_numpy` converts a possibly nested structure of `tf.data.Dataset`s
  and `tf.Tensor`s to iterables of ndarrays and ndarrays, respectively. This
  function must be run in eager mode outside tf.function.

  Args:
    dataset: a possibly nested structure of `tf.data.Dataset`s and/or
      `tf.Tensor`s.

  Returns:
    A structure matching `dataset` where `tf.data.Dataset`s are converted to
    generators of ndarrays and `tf.Tensor`s are converted to ndarrays.
  """
  if not tf.executing_eagerly():
    raise ValueError(
        "dataset_as_numpy must be run in eager mode outside tf.function")
  nested_ds = dataset
  del dataset

  # Flatten
  flat_ds = tf.nest.flatten(nested_ds)
  flat_np = []

  # Type check for Tensors and Datasets
  for ds_el in flat_ds:
    if not isinstance(ds_el, (tf.Tensor, tf.data.Dataset)):
      types = tf.nest.map_structure(type, nested_ds)
      raise ValueError("Arguments to dataset_as_numpy must be (possibly nested "
                       "structure of) tf.Tensors or tf.data.Datasets. Got: %s" %
                       types)

  for ds_el in flat_ds:
    if isinstance(ds_el, tf.Tensor):
      np_el = tf_np.asarray(ds_el)
    elif isinstance(ds_el, tf.data.Dataset):
      np_el = _eager_dataset_iterator(ds_el)
    else:
      assert False
    flat_np.append(np_el)

  return tf.nest.pack_sequence_as(nested_ds, flat_np)


# TODO(nareshmodi): Group key should change based on the set of devices that we
# are mapping over. Make it so that we assign a unique group_key for every
# unique set of devices. We don't change it every time to avoid the overhead of
# discovering the full group (though may not be problematic in the local case).
_GROUP_KEY = 1
_INSTANCE_KEY = 0
_INSTANCE_LOCK = threading.Lock()


# TODO(b/142565636): Ensure that multiple concurrent calls to a tf.function
# containing a collective op run reasonably.
def _get_instance_key():
  global _INSTANCE_KEY
  global _INSTANCE_LOCK
  with _INSTANCE_LOCK:
    _INSTANCE_KEY = _INSTANCE_KEY + 1
    return _INSTANCE_KEY


# Don't use a namedtuple since nest considers that a tuple and unflattens and
# flattens it.
class ShardedNdArray(object):
  """Wrapper over ndarray that can contain tensors on multiple devices.

    This is returned by extensions.pmap, and contains the individual tensors on
    different devices.
  """

  def __init__(self, tensors):
    """Initializes the ShardedNdArray.

    Note that the tensors should be ordered in the way the pmap producing these
    tensors is run.

    Args:
      tensors: list or tuple of eager tensors, one for each device.
    """

    if not isinstance(tensors, (list, tuple)) or not tensors:
      raise ValueError(
          "Unable to create a ShardedNdArray without a list of tensors.")
    self.tensors = tensors
    self.n_devices = len(tensors)

  def __getitem__(self, i):
    return tf_np.asarray(self.tensors[i])

  @property
  def shape(self):
    return (self.n_devices,) + self.tensors[0]._shape_tuple()  # pylint: disable=protected-access

  @property
  def dtype(self):
    return self.tensors[0].dtype


def convert_sharded_tensor_to_eager_tensor(value, *args, **kwargs):
  del args, kwargs
  # TODO(nareshmodi): Consider a collective op to gather the tensors from the
  # various devices for performance reasons.
  return tf.stack(value.tensors)


tf.register_tensor_conversion_function(ShardedNdArray,
                                       convert_sharded_tensor_to_eager_tensor)


class _PmapConfig(threading.local):
  """Simple config used to maintain state related to a current pmap call."""

  def __init__(self):
    super(_PmapConfig, self).__init__()
    self._axis_name = None
    self._devices = None

  def axis_name(self):
    return self._axis_name

  def set_axis_name(self, axis_name):
    self._axis_name = axis_name

  def devices(self):
    return self._devices

  def set_devices(self, devices):
    self._devices = devices


_pmap_config = _PmapConfig()


@contextlib.contextmanager
def pmap_config(axis_name, devices):
  """Records axis_name and devices for this context."""
  old_axis_name = _pmap_config.axis_name()
  old_devices = _pmap_config.devices()
  _pmap_config.set_axis_name(axis_name)
  _pmap_config.set_devices(devices)
  try:
    yield
  finally:
    _pmap_config.set_axis_name(old_axis_name)
    _pmap_config.set_devices(old_devices)


def _psum(tensor, axis_name=None):
  """Sum all-reduction.

  Args:
    tensor: A tensor.
    axis_name: The axis name to reduce. Must equal to that of the surrounding
      pmap.

  Returns:
    The sum of the `tensor` replicas on each participating devices.
  """
  if axis_name != _pmap_config.axis_name():
    raise ValueError("axis_name (%s) is not equal to that of the surrounding "
                     "pmap (%s)" % (axis_name, _pmap_config.axis_name()))
  devices = _pmap_config.devices()
  if devices is None:
    raise ValueError("Can't retrieve the device list from the surrounding pmap")
  tensor = tf_np.asarray(tensor)
  if tpu_devices(devices):
    # TODO(b/170895907): Remove this workaround when tpu.cross_replica_sum
    #   supports int64/float64.
    is_int64 = False
    is_float64 = False
    if tensor.dtype == np.int64:
      is_int64 = True
      tensor = tensor.astype(np.int32)
    elif tensor.dtype == np.float64:
      is_float64 = True
      tensor = tensor.astype(np.float32)
    # TODO(wangpeng): Supply the `group_assignment` argument to
    #   tpu.cross_replica_sum, calculated from `devices`.
    tensor = tf.compat.v1.tpu.cross_replica_sum(tensor)
    if is_int64:
      tensor = tf.cast(tensor, tf.int64)
    elif is_float64:
      tensor = tf.cast(tensor, tf.float64)
  else:
    tensor = tf.raw_ops.CollectiveReduce(
        input=tensor,
        group_size=len(devices),
        group_key=_GROUP_KEY,
        instance_key=_get_instance_key(),
        merge_op="Add",
        final_op="Id",
        subdiv_offsets=(0,))
  return tf_np.asarray(tensor)


def psum(tensors, axis_name=None):
  return tf.nest.map_structure(
      functools.partial(_psum, axis_name=axis_name), tensors)


# Note this is not available in the jax api, but seemed like a reasonable API
# to have.
def pmean(tensor, axis_name=None):
  """Mean all-reduction.

  Args:
    tensor: A tensor.
    axis_name: The axis name to reduce. Must equal to that of the surrounding
      pmap.

  Returns:
    The mean of the `tensor` replicas on each participating devices.
  """
  if axis_name != _pmap_config.axis_name():
    raise ValueError("axis_name (%s) is not equal to that of the surrounding "
                     "pmap (%s)" % (axis_name, _pmap_config.axis_name()))
  devices = _pmap_config.devices()
  if devices is None:
    raise ValueError("Can't retrieve the device list from the surrounding pmap")
  if tpu_devices(devices):
    # TODO(wangpeng): Implement this.
    raise ValueError("pmean for TPU is not supported yet.")
  else:
    return tf.raw_ops.CollectiveReduce(
        input=tensor,
        group_size=len(devices),
        group_key=_GROUP_KEY,
        instance_key=_get_instance_key(),
        merge_op="Add",
        final_op="Div",
        subdiv_offsets=(0,))


def _get_pmap_impl(f, devices, has_tpu):
  """This is a helper function to return the pmap impl.

  Args:
    f: a function that takes ndarrays and returns ndarrays.
    devices: a list of strings; the device list.
    has_tpu: boolean; whether `devices` contains TPU devices.

  Returns:
    A function that takes tensors and returns tensors.
  """
  if has_tpu:
    # Workaround b/121383831
    output_is_list = [False]  # Use list for mutability
    def recorder(args, kwargs, res):
      del args, kwargs
      output_is_list[0] = isinstance(res, list)
      return res
    f = _record_result_type(recorder, f)

  def tf_f(*tf_args):
    """A wrapper for `f` that takes/returns tensors."""
    np_args = _tf_to_np(tf_args)
    np_out = f(*np_args)
    return np_out

  if has_tpu:

    @tf.function(autograph=False)
    def fn(inputs):
      # TODO(wangpeng): Supply the `device_assignment` argument to
      # tpu.replicate, calculated from `devices`.
      res = tf.compat.v1.tpu.replicate(tf_f, inputs)
      # Workaround b/121383831
      if (res and isinstance(res[0], list) and len(res[0]) == 1 and
          not output_is_list[0]):
        res = [x[0] for x in res]
      return res

    return fn
  else:
    # This is run in a tf.function so that the various underlying functions can
    # be run in parallel.
    # The trace happens on the client, so any devices should not depend on any
    # side effects.

    jit_tf_f = tf.function(tf_f, autograph=False)

    @tf.function(autograph=False)
    def fn(all_per_device_args):
      """Multi-device function with calls placed on the correct device."""

      results = []
      for per_device_args, device in zip(all_per_device_args, devices):
        with tf.device(device):
          results.append(jit_tf_f(*per_device_args))
      return results

    return fn


def pmap(f, axis_name=None, devices=None):
  """Transforms a function into a multi-device function.

  The semantics are similar to JAX's pmap.

  Args:
    f: The function to be converted.
    axis_name: Used for nested pmap, which is not supported yet.
    devices: The devices over which the returned function will run.

  Returns:
    A function that runs the underlying function `f` on `devices`. Its arguments
    can be `ShardedNdArray`s, tensors or other Python objects, and its return
    values are all `ShardedNdArray`s. If an input is a tensor, the length of its
    first dimension must equal the number of devices, and the tensor will be
    splitted along its first dimension among the devices. If an input is an
    unknown Python object, it will be replicated among the devices.
  """
  if devices is None:
    devices = accelerators()
  if not isinstance(devices, (list, tuple)):
    raise ValueError("Must pass a list or tuple of devices")
  num_devices = len(devices)
  if not num_devices:
    raise ValueError("There must be at least 1 device")
  has_tpu = bool(tpu_devices(devices))

  pmap_fn = _get_pmap_impl(f, devices, has_tpu)

  def wrapper(*args):
    """Wrapper that wraps/unwraps args, retvals, and runs the function."""
    if _pmap_config.devices() is not None:
      raise ValueError("Found a surrounding pmap. Nested pmap is not supported "
                       "yet.")
    # TODO(wangpeng): Maybe we should use `asarray` to convert everything
    # to ndarray first.

    flattened_input_args = tf.nest.flatten(args)
    flattened_per_device_args = [[] for _ in devices]
    for arg in flattened_input_args:
      if isinstance(arg, tf.Tensor):
        # TODO(nareshmodi): Try and use the dynamic shape instead.
        if (not arg.shape.rank) or arg.shape[0] != len(devices):
          # TODO(nareshmodi): Fix this restriction
          raise ValueError(
              "Input tensors need to have a first dimension equal to "
              "the number of devices; got tensor of shape %s and %s devices" %
              (arg.shape, len(devices)))
        # NOTE: Alternatively use tf.split, and place the split tensors on the
        # appropriate device. The best solution for this is to have an API that
        # splits a tensor across devices.
        for j, device in enumerate(devices):
          updated_arg = tf.gather(arg, j)
          # TODO(wangpeng): Investigate whether we need a tf.identity for TPU.
          if not has_tpu:
            with tf.device(device):
              updated_arg = tf.identity(updated_arg)
          flattened_per_device_args[j].append(updated_arg)
      elif isinstance(arg, ShardedNdArray):
        for device_args, tensor in zip(flattened_per_device_args, arg.tensors):
          device_args.append(tensor)
      else:
        for device_args in flattened_per_device_args:
          device_args.append(arg)

    all_per_device_args = [
        tf.nest.pack_sequence_as(args, device_args)
        for device_args in flattened_per_device_args
    ]

    with pmap_config(axis_name, devices):
      results = pmap_fn(all_per_device_args)

    # Rewrap things. This can probably be written better.
    flattened_results = [tf.nest.flatten(result) for result in results]
    final_tree = []

    # TODO(nareshmodi): assert all items in flattened_results have the same
    # structures

    for i in range(len(flattened_results[0])):
      tensors = []
      for j, device in enumerate(devices):
        assert isinstance(
            flattened_results[j][i],
            tf.Tensor), ("currently only tensor return items are supported")
        tensors.append(flattened_results[j][i])
      final_tree.append(ShardedNdArray(tensors))

    return tf.nest.pack_sequence_as(results[0], final_tree)

  return wrapper


def find_devices(device_type, devices=None):
  if not devices:
    devices = [d.name for d in tf.config.experimental.list_logical_devices()]
  devices = [(d, tf.DeviceSpec.from_string(d)) for d in devices]
  results = [name for name, d in devices if d.device_type == device_type]
  return results


def tpu_devices(devices=None):
  """Gets TPU devices out of `devices`.

  Args:
    devices: A device list (as a list of strings). If None, the list of all
      available devices will be used for it.

  Returns:
    Those in `devices` that are TPUs.
  """
  return find_devices("TPU", devices)


def gpu_devices(devices=None):
  """Gets GPU devices out of `devices`.

  Args:
    devices: A device list (as a list of strings). If None, the list of all
      available devices will be used for it.

  Returns:
    Those in `devices` that are GPUs.
  """
  return find_devices("GPU", devices)


def accelerators(devices=None):
  return tpu_devices(devices) or gpu_devices(devices)


def _tree_broadcast(to, s):
  """Broadcasts `s` to the nested structure `to`."""
  if not isinstance(to, (list, tuple, dict)):
    if not isinstance(s, (int, type(None))):
      raise ValueError
    return s
  if isinstance(s, (int, type(None))):
    return tf.nest.map_structure(lambda x: s, to)
  if isinstance(to, (list, tuple)):
    if len(to) != len(s):
      raise ValueError
    new_s = [_tree_broadcast(x, y) for x, y in zip(to, s)]
    if isinstance(to, tuple):
      new_s = tuple(new_s)
    return new_s
  elif isinstance(to, dict):
    return {k: _tree_broadcast(to[k], s[k]) for k in to}
  else:
    raise TypeError("Unsupported type %s" % type(to))


def vmap(f, in_axes=0, out_axes=0):
  """Returns a function that maps `f` over first dimension of inputs."""
  in_axes_flat = tf.nest.flatten(in_axes)
  if not all(isinstance(l, (type(None), int))
             for l in in_axes_flat):
    raise TypeError(
        "vmap in_axes must be an int, None, or (nested) container with "
        "those types as leaves, but got {}.".format(in_axes))
  if all(isinstance(l, type(None)) for l in in_axes_flat):
    raise ValueError("vmap must have at least one non-None value in in_axes")

  out_axes_flat = tf.nest.flatten(out_axes)
  if not all(isinstance(l, (type(None), int))
             for l in out_axes_flat):
    raise TypeError(
        "vmap out_axes must be an int, None, or (nested) container with "
        "those types as leaves, but got {}.".format(out_axes))

  def _f(*args):
    flat_args = tf.nest.flatten(args)
    try:
      f_in_axes = _tree_broadcast(args, in_axes)
    except ValueError:
      six.reraise(
          ValueError,
          ValueError(
              "vmap in_axes specification must be a tree prefix of the "
              r"corresponding value, got specification %s for value tree %s" % (
                  in_axes, args)),
          sys.exc_info()[2])
    f_in_axes_flat = tf.nest.flatten(f_in_axes)

    def tf_f(tf_args):
      """Function passed to tf.vectorized_map call."""
      # Note that unbatched arguments are not passed to tf_f. Here we fill thos
      # arguments back before calling `f`.
      tf_flat_args = []
      j = 0
      for arg, axis in zip(flat_args, f_in_axes_flat):
        if axis is None:
          tf_flat_args.append(arg)
        else:
          tf_flat_args.append(tf_args[j])
          j += 1
      unbatched_args = tf.nest.pack_sequence_as(args, tf_flat_args)
      return f(*unbatched_args)

    # Constructs arguments to pass to `tf_f`.
    # Unbatch arguments are skipped. Arguments with non-zero axis are
    # transposed.
    tf_args = []
    for arg, axis in zip(flat_args, f_in_axes_flat):
      if axis is None:
        continue
      arg = tf_np.asarray(arg)
      if axis != 0:
        arg = tf_np.moveaxis(arg, axis, 0)
      tf_args.append(arg)
    # TODO(agarwal): consider creating a tf.function outside of _f and reusing
    # that to avoid overheads of re-vectorizing the code when running eagerly.
    outputs = tf.vectorized_map(tf_f, tf_args)
    try:
      f_out_axes = _tree_broadcast(outputs, out_axes)
    except ValueError:
      six.reraise(
          ValueError,
          ValueError(
              "vmap out_axes specification must be a tree prefix of the "
              r"corresponding value, got specification %s for value tree %s" % (
                  out_axes, outputs)),
          sys.exc_info()[2])

    def map_output(x, axis):
      """Maps output of tf.vectorized_map to the final output."""
      x = tf_np.asarray(x)
      if axis is None:
        # Note that `tf.vectorized_map always batches the outputs.
        # Here we unbatch it again.
        return x[0, ...]
      elif axis == 0:
        return x
      else:
        # Need to transpose the output.
        return tf_np.moveaxis(x, 0, axis)
    new_outputs = [map_output(output, axis) for output, axis in zip(
        tf.nest.flatten(outputs), tf.nest.flatten(f_out_axes))]
    return tf.nest.pack_sequence_as(outputs, new_outputs)

  return _f



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Normalization helpers."""

import gin
import numpy as np

from trax import fastmath
from trax import layers as tl


def running_mean_init(shape, fill_value=0):
  return (np.full(shape, fill_value), np.array(0))


def running_mean_update(x, state):
  (mean, n) = state
  mean = n.astype(np.float32) / (n + 1) * mean + x / (n + 1)
  return (mean, n + 1)


def running_mean_get_mean(state):
  (mean, _) = state
  return mean


def running_mean_get_count(state):
  (_, count) = state
  return count


def running_mean_and_variance_init(shape):
  mean_state = running_mean_init(shape, fill_value=0.0)
  var_state = running_mean_init(shape, fill_value=1.0)
  return (mean_state, var_state)


def running_mean_and_variance_update(x, state):
  (mean_state, var_state) = state
  old_mean = running_mean_get_mean(mean_state)
  mean_state = running_mean_update(x, mean_state)
  new_mean = running_mean_get_mean(mean_state)

  var_state = running_mean_update((x - new_mean) * (x - old_mean), var_state)

  return (mean_state, var_state)


def running_mean_and_variance_get_mean(state):
  (mean_state, _) = state
  return running_mean_get_mean(mean_state)


def running_mean_and_variance_get_count(state):
  (mean_state, _) = state
  return running_mean_get_count(mean_state)


def running_mean_and_variance_get_variance(state):
  (_, var_state) = state
  return running_mean_get_mean(var_state)


@gin.configurable(denylist=['mode'])
class Normalize(tl.Layer):
  """Numerically stable normalization layer."""

  def __init__(self, sample_limit=float('+inf'), epsilon=1e-5, mode='train'):
    super().__init__()
    self._sample_limit = sample_limit
    self._epsilon = epsilon
    self._mode = mode

  def init_weights_and_state(self, input_signature):
    self.state = running_mean_and_variance_init(input_signature.shape[2:])

  def forward(self, inputs):
    state = self.state
    observations = inputs
    if self._mode == 'collect':
      # Accumulate statistics only in the collect mode, i.e. when collecting
      # data using the agent.
      for observation in observations[:, -1]:  # (batch_size, time, ...)
        # Update statistics for each observation separately for simplicity.
        # Currently during data collection the batch size is 1 anyway.
        count = running_mean_and_variance_get_count(state)
        state = fastmath.cond(
            count < self._sample_limit,
            true_operand=(observation, state),
            true_fun=lambda args: running_mean_and_variance_update(*args),
            false_operand=None,
            false_fun=lambda _: state,
        )

    mean = running_mean_and_variance_get_mean(state)
    var = running_mean_and_variance_get_variance(state)
    norm_observations = (observations - mean) / (var ** 0.5 + self._epsilon)
    self.state = state
    return norm_observations


@gin.configurable(denylist=['mode'])
def LayerNormSquash(mode, width=128):  # pylint: disable=invalid-name
  """Dense-LayerNorm-Tanh normalizer inspired by ACME."""
  # https://github.com/deepmind/acme/blob/master/acme/jax/networks/continuous.py#L34
  del mode
  return tl.Serial([
      tl.Dense(width),
      tl.LayerNorm(),
      tl.Tanh(),
  ])



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for testing."""

import copy
import functools

import numpy as np

from trax import fastmath
from trax import layers as tl
from trax import shapes


def test_eval_is_deterministic(inp, model_fn, message=''):
  """Utility method for testing if eval mode is deterministic.

  Args:
    inp: input fed to the model. It can be a tensor, or a tuple of tensors.
    model_fn: function creating a model after calling with `mode` argument.
    message: Optional message to show when outputs of eval/predict mode don't
      match.
  """
  with fastmath.use_backend(fastmath.Backend.JAX):
    model_eval1 = model_fn(mode='eval')
    model_eval2 = model_fn(mode='eval')

    input_signature = shapes.signature(inp)
    model_eval1.init(input_signature)
    model_eval2.init(input_signature)
    model_eval1.save_to_file('/tmp/unique_weights')
    model_eval2.init_from_file('/tmp/unique_weights', weights_only=True,
                               input_signature=input_signature)

    rng = fastmath.random.get_prng(0)
    output_eval1 = model_eval1(inp, rng=rng)
    if not isinstance(output_eval1, (tuple, list)):
      # We will automatically check each and every tensor returned.
      output_eval1 = [output_eval1]

    output_eval2 = model_eval2(inp, rng=rng)
    if not isinstance(output_eval2, (tuple, list)):
      # We will automatically check each and every tensor returned.
      output_eval2 = [output_eval2]

    np.testing.assert_equal(len(output_eval1), len(output_eval2))
    for out1, out2 in zip(output_eval1, output_eval2):
      np.testing.assert_array_almost_equal(
          out1,
          out2,
          decimal=5,
          err_msg='Non-deterministic.{}'.format(message))


def test_eval_equals_predict(inp, model_fn, seq_axis=1, seq_tensor=None,
                             init_tokens=3, message=''):
  """Utility method for testing equivalence of predict and eval modes.

  Args:
    inp: input fed to the model. It can be a tensor, or a tuple of tensors.
    model_fn: function creating a model after calling with `mode` argument.
    seq_axis: axis of sequence_length. In predict mode we iterate over this
      axis. By default `1`, which is 2nd dimension.
    seq_tensor: if `inp` is a tuple, `seq_tensor` is an index of an input tensor
      in this tuple on which we iterate the sequence.
    init_tokens: how many tokens should be passed to the first `predict` call.
    message: Optional message to show when outputs of eval/predict mode don't
      match.
  """
  with fastmath.use_backend(fastmath.Backend.JAX):
    model_eval = model_fn(mode='eval')
    model_predict = model_fn(mode='predict')

    input_signature = shapes.signature(inp)
    model_eval.init(input_signature)
    model_predict.init(input_signature)
    model_eval.save_to_file('/tmp/unique_weights')
    model_predict.init_from_file('/tmp/unique_weights', weights_only=True,
                                 input_signature=input_signature)

    rng = fastmath.random.get_prng(0)
    output_eval = model_eval(inp, rng=rng)
    if not isinstance(output_eval, (tuple, list)):
      # We will automatically check each and every tensor returned.
      output_eval = [output_eval]

    if seq_tensor is None:
      length = inp.shape[seq_axis]
    else:
      length = inp[seq_tensor].shape[seq_axis]

    assert length >= init_tokens + 2  # Required to properly test predict mode.
    indices_list = [(0, init_tokens)] + [(i, i+1)
                                         for i in range(init_tokens, length)]

    for indices in indices_list:
      start, end = indices
      if seq_tensor is None:
        new_inp = inp.take(indices=np.arange(start, end), axis=seq_axis)
      else:
        new_inp = list(inp)
        new_inp[seq_tensor] = new_inp[seq_tensor].take(
            indices=np.arange(start, end), axis=seq_axis)

      output_predict = model_predict(new_inp, rng=rng)
      if not isinstance(output_predict, (tuple, list)):
        # We will automatically check each and every tensor returned.
        output_predict = [output_predict]

      np.testing.assert_equal(len(output_predict), len(output_eval))
      for outp, oute in zip(output_predict, output_eval):
        np.testing.assert_array_almost_equal(
            oute.take(indices=np.arange(start, end), axis=seq_axis),
            outp.take(indices=np.arange(0, end-start), axis=seq_axis),
            decimal=5,
            err_msg='Error on element {} out of {}.{}'.format(indices, length,
                                                              message))


def test_eval_equals_predict_configs(inp, model_fn, configs, seq_axis=1,
                                     seq_tensor=None, message=''):
  """Utility method for testing equivalence of predict and eval modes.

  This function iterates over a list of dictionaries `confis`, and runs the test
  on models with each configuration.

  Args:
    inp: input fed to the model. It can be a tensor, or a tuple of tensors.
    model_fn: function creating a model after calling with `mode` argument.
    configs: List of dictionaries, which contain configs to be fed into
      `model_fn`.
    seq_axis: axis of sequence_length. In predict mode we iterate over this
      axis. By default `1`, which is 2nd dimension.
    seq_tensor: if `inp` is a tuple, `seq_tensor` is an index of an input tensor
      in this tuple on which we iterate the sequence.
    message: Optional message to show when outputs of eval/predict mode don't
      match.
  """
  for config in configs:
    model_fn_configured = functools.partial(model_fn, **config)
    test_eval_equals_predict(inp, model_fn_configured, seq_axis=seq_axis,
                             seq_tensor=seq_tensor,
                             message=' Config: {}.{}'.format(config, message))


def test_eval_equals_predict_discrete(
    model_fn, vocab_size=10, length=5, batch_size=3
):
  """Tests the equivalence of eval and predict modes for discrete models."""
  with fastmath.use_backend(fastmath.Backend.JAX):
    model_slow = model_fn(mode='eval', vocab_size=vocab_size)
    model_fast = model_fn(mode='predict', vocab_size=vocab_size)
    rng = fastmath.random.get_prng(0)
    input_signature = shapes.ShapeDtype((batch_size, 1), np.int32)
    # Given the same rng, both models initialize with the same parameters.
    model_slow.init(input_signature, rng)
    model_fast.init(input_signature, rng)

    buf = np.zeros((batch_size, length), dtype=np.int32)
    next_sym = np.zeros((batch_size, 1), dtype=np.int32)

    for index in range(length):
      logits_slow = model_slow(buf, rng=rng)
      logits_fast = model_fast(next_sym, rng=rng)
      np.testing.assert_array_almost_equal(
          logits_slow[:, index, :], logits_fast[:, 0, :],
          decimal=5,
      )
      next_sym = np.random.randint(vocab_size, size=(batch_size, 1))
      buf[:, index] = next_sym[:, 0]


class MockTransformerLM(tl.Layer):
  r"""Mock TransformerLM for testing autoregressive sampling routines.

  Mimics the behavior of a perfectly-trained, deterministic TransformerLM.
  Allows to specify the \sigma^* -> \sigma function implemented by the model
  and to make assertions about the input sequence passed to the model.

  Supports two modes: stateful "predict" for fast inference, and stateless
  non-"predict" ("train", "eval" etc).

  Useful for testing any logic that relies on autoregressive sampling, as it
  removes the additional layer of complexity related to training a model or
  maintaining a pretrained one. Makes the tests run MUCH faster.

  Does not support acceleration. Do not wrap in tl.Accelerate().
  """

  def __init__(self, sequence_fn, mode, vocab_size):
    super().__init__()

    self._sequence_fn = sequence_fn
    self._mode = mode
    self._vocab_size = vocab_size

    self._prediction_buffers = None

  @property
  def state(self):
    return copy.deepcopy(self._prediction_buffers)

  @state.setter
  def state(self, state):
    self._prediction_buffers = copy.deepcopy(state)

  def _output_symbol_predict(self, input_symbols, prediction_buffer):
    prediction_buffer.extend(input_symbols)
    output_symbol = self._sequence_fn(np.array(prediction_buffer))
    return np.array([output_symbol])

  def _output_symbols_eval(self, input_symbols, prediction_buffer):
    del prediction_buffer

    # Add a leading 0 token to imitate ShiftRight.
    input_symbols = np.concatenate(([0], input_symbols))

    # Call sequence_fn repeatedly along the input sequence.
    return np.array([
        self._sequence_fn(input_symbols[:end])
        for end in range(1, len(input_symbols))
    ])

  def _symbols_to_logits(self, symbols):
    # Assert that symbols are discrete.
    assert np.issubdtype(symbols.dtype, np.integer)
    # Assert that 0 <= symbols < vocab_size.
    np.testing.assert_array_less(-1, symbols)
    np.testing.assert_array_less(symbols, self._vocab_size)

    # Return almost-determinisitc logits:
    # e^1000 / (e^1000 + vocab_size) ~= 1
    return tl.one_hot(symbols, n_categories=self._vocab_size) * 1000.0

  def __call__(self, inputs, rng=None):
    del rng

    assert inputs.ndim == 2, (
        'The input sequences should have exactly two axes.'
    )

    if self._prediction_buffers is None:
      # Initialize the buffer.
      batch_size = inputs.shape[0]
      # [[]] * batch_size would create multiple references to the same
      # list, and we want separate lists.
      self._prediction_buffers = [[] for _ in range(batch_size)]

    if self._mode == 'predict':
      output_fn = self._output_symbol_predict
    else:
      output_fn = self._output_symbols_eval

    # Calculate the output separately for each sequence in the batch.
    output_symbols = np.array([
        output_fn(input_seq, pred_buffer)
        for (input_seq, pred_buffer) in zip(
            inputs, self._prediction_buffers
        )
    ])
    return self._symbols_to_logits(output_symbols)

  def assert_prediction_buffers_equal(self, expected_buffers):
    if self._prediction_buffers is None:
      batch_size = expected_buffers.shape[0]
      actual_buffers = np.empty((batch_size, 0))
    else:
      actual_buffers = np.array(self._prediction_buffers)

    np.testing.assert_array_equal(actual_buffers, expected_buffers)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Trax accelerated math operations for fast computing on GPUs and TPUs.

Import these operations directly from fastmath and import fastmath.numpy as np::

    from trax import fastmath
    from trax.fastmath import numpy as np

    x = np.array([1.0, 2.0])  # Use like numpy.
    y = np.exp(x)  # Common numpy ops are available and accelerated.
    z = fastmath.logsumexp(y)  # Special operations available from fastmath.

Trax uses either TensorFlow 2 or JAX as backend for accelerating operations.
You can select which one to use (e.g., for  debugging) with `use_backend`.
"""

import contextlib
import enum

import gin
from trax.fastmath.jax import JAX_BACKEND
from trax.fastmath.numpy import NUMPY_BACKEND
from trax.fastmath.tf import TF_BACKEND


@enum.unique
class Backend(enum.Enum):
  JAX = 'jax'
  TFNP = 'tensorflow-numpy'
  NUMPY = 'numpy'


# For numpy and random modules, we need to call "backend()" lazily, only when
# the function is called -- so that it can be set by gin configs.
# (Otherwise, backend() is called on import before gin-config is parsed.)
# To do that, we make objects to encapsulated these modules.


# A class that just forwards attribute accesses to backend's numpy object.
class NumpyBackend:
  """Numpy functions accelerated to run on GPUs and TPUs. Use like numpy."""

  def __getattr__(self, attr):
    return getattr(backend()['np'], attr)

numpy = NumpyBackend()


class RandomBackend:
  """Backend providing random functions."""

  def get_prng(self, seed):
    return backend()['random_get_prng'](seed)

  def split(self, prng, num=2):
    return backend()['random_split'](prng, num)

  def fold_in(self, rng, data):
    return backend()['random_fold_in'](rng, data)

  def uniform(self, *args, **kwargs):
    return backend()['random_uniform'](*args, **kwargs)

  def randint(self, *args, **kwargs):
    return backend()['random_randint'](*args, **kwargs)

  def normal(self, *args, **kwargs):
    return backend()['random_normal'](*args, **kwargs)

  def bernoulli(self, *args, **kwargs):
    return backend()['random_bernoulli'](*args, **kwargs)


random = RandomBackend()


def logsumexp(*args, **kwargs):
  """Computes the log of the sum of exponentials of input elements."""
  return backend()['logsumexp'](*args, **kwargs)


def expit(*args, **kwargs):
  """Computes the expit (sigmoid) function."""
  return backend()['expit'](*args, **kwargs)


def sigmoid(*args, **kwargs):
  """Computes the sigmoid (expit) function."""
  return backend()['expit'](*args, **kwargs)


def erf(*args, **kwargs):
  """Computes the erf function."""
  return backend()['erf'](*args, **kwargs)


def conv(*args, **kwargs):
  """Computes a generalized convolution."""
  return backend()['conv'](*args, **kwargs)


def avg_pool(*args, **kwargs):
  """Average pooling."""
  return backend()['avg_pool'](*args, **kwargs)


def max_pool(*args, **kwargs):
  """Max pooling."""
  return backend()['max_pool'](*args, **kwargs)


def sum_pool(*args, **kwargs):
  """Sum pooling."""
  return backend()['sum_pool'](*args, **kwargs)


def top_k(*args, **kwargs):
  """Top k."""
  return backend()['top_k'](*args, **kwargs)


def sort_key_val(*args, **kwargs):
  """Sorts keys along dimension and applies same permutation to values."""
  return backend()['sort_key_val'](*args, **kwargs)


def scan(*args, **kwargs):
  """Scan to make recurrent functions run faster on accelerators."""
  return backend()['scan'](*args, **kwargs)


def map(*args, **kwargs):  # pylint: disable=redefined-builtin
  """Map a function over leading array axes."""
  return backend()['map'](*args, **kwargs)


def fori_loop(lower, upper, body_fn, init_val):
  """Loop from `lower` to `upper` running `body_fn` starting from `init_val`.

  The semantics of `fori_loop` is as follows::

    def fori_loop(lower, upper, body_fn, init_val):
      val = init_val
      for i in range(lower, upper):
        val = body_fn(i, val)
      return val

  Args:
    lower: an integer representing the loop index lower bound (inclusive)
    upper: an integer representing the loop index upper bound (exclusive)
    body_fn: function of type `(int, a) -> a`.
    init_val: initial loop carry value of type `a`.

  Returns:
    Loop value from the final iteration.
  """
  if 'fori_loop' in backend():
    return backend()['fori_loop'](lower, upper, body_fn, init_val)
  # Use scan otherwise.
  def scanned_fn(loop_carry, _):
    i, x = loop_carry
    return (i + 1, body_fn(i, x)), None
  (_, result), _ = scan(
      scanned_fn, (lower, init_val), None, length=upper - lower)
  return result


def remat(*args, **kwargs):
  """Recompute everything in the backward pass to same memory."""
  return backend()['remat'](*args, **kwargs)


def cond(*args, **kwargs):
  """Conditional computation to run on accelerators."""
  return backend()['cond'](*args, **kwargs)


def lt(*args, **kwargs):
  """Less-than function for backends that do not override <."""
  return backend()['lt'](*args, **kwargs)


def index_update(*args, **kwargs):
  return backend()['index_update'](*args, **kwargs)


def index_add(*args, **kwargs):
  return backend()['index_add'](*args, **kwargs)


def index_min(*args, **kwargs):
  return backend()['index_min'](*args, **kwargs)


def index_max(*args, **kwargs):
  return backend()['index_max'](*args, **kwargs)


def dynamic_slice(*args, **kwargs):
  return backend()['dynamic_slice'](*args, **kwargs)


def dynamic_slice_in_dim(*args, **kwargs):
  return backend()['dynamic_slice_in_dim'](*args, **kwargs)


def dynamic_update_slice(*args, **kwargs):
  return backend()['dynamic_update_slice'](*args, **kwargs)


def dynamic_update_slice_in_dim(*args, **kwargs):
  return backend()['dynamic_update_slice_in_dim'](*args, **kwargs)


def stop_gradient(*args, **kwargs):
  """Identity on the forward pass but 0 (no gradient) on the backward pass."""
  return backend()['stop_gradient'](*args, **kwargs)


_disable_jit = False


def jit(*args, **kwargs):
  """Just-In-Time compiles the given function for use on accelerators."""
  global _disable_jit
  if _disable_jit:
    return args[0]  # jit(f, **unused_now_jit_kwargs) = f
  return backend()['jit'](*args, **kwargs)


def disable_jit():
  """Disables JIT-compilation; helpful for debugging."""
  global _disable_jit
  _disable_jit = True


def vmap(*args, **kwargs):
  """Vectorizes the specified function (returns a function)."""
  return backend()['vmap'](*args, **kwargs)


def grad(*args, **kwargs):
  """Computes the gradient of the specified function (returns a function)."""
  return backend()['grad'](*args, **kwargs)


def value_and_grad(*args, **kwargs):
  """Computes the gradient of the specified function together with the value."""
  if 'value_and_grad' in backend():
    return backend()['value_and_grad'](*args, **kwargs)
  grad_fn = grad(*args, **kwargs)
  fn = args[0]
  has_aux = False
  if has_aux in kwargs:
    has_aux = kwargs['has_aux']
  if not has_aux:
    def val_and_grad(*fn_args, **fn_kwargs):
      return fn(*fn_args, **fn_kwargs), grad_fn(*fn_args, **fn_kwargs)
    return val_and_grad
  def val_and_grad_aux(*fn_args, **fn_kwargs):
    g, aux = grad_fn(*fn_args, **fn_kwargs)
    res, _ = fn(*fn_args, **fn_kwargs)
    return (res, aux), g
  return val_and_grad_aux


def vjp(*args, **kwargs):
  """Computes the vector-Jacobian product for the specified function."""
  return backend()['vjp'](*args, **kwargs)


def custom_grad(*args, **kwargs):
  """Set a custom gradient computation (override the default) for a function."""
  return backend()['custom_grad'](*args, **kwargs)


def custom_vjp(f, f_fwd, f_bwd, nondiff_argnums=()):
  """Set a custom vjp computation (override the default) for a function."""
  # Call backend custom_vjp if it exists.
  # TODO(lukaszkaiser): unify the APIs and remove nondiff_argnums altogether.
  if 'custom_vjp' in backend():
    return backend()['custom_vjp'](f, f_fwd, f_bwd)

  # Check that nondiff_argnums is (0, 1, ..., N) for some N.
  # Currently we only support nondiff_argnums at the front.
  counter = -1
  for i in nondiff_argnums:
    counter += 1
    if i != counter:
      raise ValueError('Currently we only support custom_vjps with all nondiff'
                       '_argnums up front, like (0,) or (0, 1) but not (1,) or'
                       ' (1, 2). Found: %s' % str(nondiff_argnums))

  # Use custom_grad.
  if counter == -1:  # no non-diff args
    def f_vjp(*args):
      out, residual = f_fwd(*args)
      def vjpfn(g):
        return f_bwd(residual, g)
      return out, vjpfn
    return backend()['custom_grad'](f_vjp, f)

  # Handle non-diff args by closure.
  def f_joint(*args):
    """This function takes all args, first counter+1 are non-diff ones."""
    nondiff_args = list(args[:counter+1])
    def f_diff(*diff_args):  # Takes only diff args, will define custom grad.
      args = nondiff_args + list(diff_args)
      return f(*args)
    def f_vjp(*diff_args):  # Custom VJP for diff args.
      args = nondiff_args + list(diff_args)
      out, residual = f_fwd(*args)
      def vjpfn(g):
        bwd_args = [residual, g]
        res = f_bwd(*bwd_args)
        return res[counter+1:]
      return out, vjpfn
    # This is the function taking only diff args with custom vjp.
    f_diff_vjp = backend()['custom_grad'](f_vjp, f_diff)
    # Call it on the diff args.
    return f_diff_vjp(*args[counter+1:])
  return f_joint


def pmap(*args, **kwargs):
  """Parallel-map to apply a function on multiple accelerators in parallel."""
  return backend()['pmap'](*args, **kwargs)


def psum(*args, **kwargs):
  """Parallel-sum to use within a pmap'd function for aggregation."""
  return backend()['psum'](*args, **kwargs)


def abstract_eval(*args, **kwargs):
  """Evaluates function just on signatures of parameters, return signatures."""
  return backend()['abstract_eval'](*args, **kwargs)


def dataset_as_numpy(*args, **kwargs):
  """Convert a tf.data.Dataset to a stream of numpy arrays."""
  if 'dataset_as_numpy' in backend():
    return backend()['dataset_as_numpy'](*args, **kwargs)
  return JAX_BACKEND['dataset_as_numpy'](*args, **kwargs)


def global_device_count(*args, **kwargs):
  """Return the number of accelerators (GPUs or TPUs) in all hosts."""
  return backend()['global_device_count'](*args, **kwargs)


def local_device_count(*args, **kwargs):
  """Return the number of accelerators (GPUs or TPUs) available on this host."""
  return backend()['local_device_count'](*args, **kwargs)


# Backend selection functions.

override_backend = None
default_backend = None
_backend_dict = {
    Backend.JAX: JAX_BACKEND,
    Backend.NUMPY: NUMPY_BACKEND,
    Backend.TFNP: TF_BACKEND,
}


def _assert_valid_backend_name(name):
  for backend_ in Backend:
    if backend_.value == name:
      return
  raise ValueError(f'No backend with name {name}')


def set_backend(name):
  """Sets the default backend to use in Trax."""
  if name:
    _assert_valid_backend_name(name)
  global default_backend
  default_backend = name


def _get_backend_from_string(name_str):
  # name is a string.
  for backend_ in Backend:
    if backend_.value == name_str:
      return _backend_dict[backend_]
  return JAX_BACKEND


@gin.configurable
def backend(name='jax'):
  """Returns the backend used to provide fastmath ops ('tf' or 'jax')."""
  if override_backend:
    return _get_backend_from_string(override_backend)

  if default_backend:
    return _get_backend_from_string(default_backend)

  if isinstance(name, Backend):
    return _backend_dict[name]

  # name is a string.
  return _get_backend_from_string(name)


@contextlib.contextmanager
def use_backend(name):
  """Call fastmath functions with a specified backend."""
  if isinstance(name, Backend):
    name = name.value

  _assert_valid_backend_name(name)
  global override_backend
  prev_name_or_backend = override_backend
  override_backend = name
  # Run the decorated function in try-finally in case it throws, e.g. for tests.
  try:
    yield
  finally:
    override_backend = prev_name_or_backend


def backend_name():
  """Returns the name of the backend currently in use ('tf' or 'jax')."""
  return backend()['name']


def is_backend(backend_):
  return backend()['name'] == backend_.value



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Prediction binary for the Drop task.

Binary that loads a checkpoint and runs inference on selected problems
from the Drop dataset. For more details about Drop see
https://arxiv.org/pdf/1903.00161.pdf.
"""

import json
import os
import re
import time

from absl import app as absl_app
from absl import flags
import gin
import jax
import numpy as np
from seqio import vocabularies as t5_spc_vocab
from t5 import data
import tensorflow as tf
from trax import data as trax_data
from trax import layers as tl
from trax import shapes
from trax.supervised import decoding


FLAGS = flags.FLAGS

flags.DEFINE_string('checkpoint_dir', '',
                    'Path to model checkpoint.')
flags.DEFINE_integer('max_answer_len', 1024,
                     'Maximum length of answers to produce.')
flags.DEFINE_integer('batch_size', 1, 'Batch size for eval.')
flags.DEFINE_integer('num_examples', 1, 'Number of examples to infer.')
flags.DEFINE_integer('n_hashes', None,
                     'n_hashes parameter to override in attentions.')
flags.DEFINE_integer('example_repetitions', 1,
                     'How many times to infer an example.')
flags.DEFINE_bool('use_eval_mode', False,
                  'If True, use the slower but easier to debug eval mode.')
flags.DEFINE_bool('use_eval_set', False,
                  'If True, use eval set for evaluation.')
flags.DEFINE_bool(
    'use_beam_search', False,
    'If True, use beam search, otherwise use autoregresive sampling.')
flags.DEFINE_float('autoregressive_sample_temp', 1,
                   'The temperature for autoregressive sampling.')
flags.DEFINE_integer('n_beams', 4, 'How many beams to use in beam search.')
flags.DEFINE_string(
    'output_dir', '', 'Path to the output directory where articles, abstracts, '
    'and predictions would be stored.')
flags.DEFINE_integer('starting_example', 0,
                     'Example index for starting decoding.')
flags.DEFINE_integer('reload_after', 1000,
                     'Reload checkpoint after reload_after examples.')
flags.DEFINE_multi_string('config_file', None,
                          'Configuration file with parameters (.gin).')


def _check_exists(file_path):
  if not tf.io.gfile.exists(file_path):
    print('No such file: %s' % file_path, flush=True)
    exit(1)


def multiply_examples(example):
  for i in range(FLAGS.example_repetitions):
    yield i, example


def prepare_model(model_file, batch_size=1):
  """Prepare the model."""
  mode = 'eval' if FLAGS.use_eval_mode else 'predict'
  print('Initializing the model in %s mode.' % mode, flush=True)

  # Read the model name from the gin file
  model_reference = gin.query_parameter(
      'trax.supervised.trainer_lib.train.model')
  model = model_reference.scoped_configurable_fn(mode=mode)

  dec_len = 32 if FLAGS.use_eval_mode else 1
  batch_size_pd = max(1, batch_size // jax.local_device_count())
  shape11 = shapes.ShapeDtype((batch_size_pd, dec_len), dtype=np.int32)
  # shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
  model.init_from_file(
      model_file, weights_only=True, input_signature=(shape11, shape11))
  model = tl.Accelerate(model)

  initial_state = model.state
  vocab = t5_spc_vocab.SentencePieceVocabulary(data.DEFAULT_SPM_PATH)

  return vocab, model, initial_state


def is_number(s):
  try:
    float(s)
    return True
  except ValueError:
    return False


def main(argv):
  if len(argv) > 1:
    raise absl_app.UsageError('Too many command-line arguments.')
  if not FLAGS.output_dir:
    raise absl_app.UsageError('--output_dir needs to be provided.')

  tf.compat.v1.enable_eager_execution()

  # Check that checkpoint_dir is correct: should contain model.pkl.gz file.
  model_file = os.path.join(FLAGS.checkpoint_dir, 'model.pkl.gz')
  _check_exists(model_file)

  gin.parse_config_file(os.path.join(FLAGS.checkpoint_dir, 'config.gin'))
  # Batching on our own because of possible repetitions of examples.
  gin.bind_parameter('data.Batch.batch_size', 1)
  if FLAGS.n_hashes is not None:
    gin.bind_parameter('LSHSelfAttention.n_hashes', FLAGS.n_hashes)
    gin.bind_parameter('ref2_encoder/LSHSelfAttention.n_hashes', FLAGS.n_hashes)

  vocab, model, initial_state = prepare_model(model_file, FLAGS.batch_size)

  host_id, host_count = jax.host_id(), jax.host_count()
  print('Running on host %d out of %d.' % (host_id, host_count))

  example_count = 0
  start_time = time.time()

  # Creates all intermediate directories if they do not exist
  tf.io.gfile.makedirs(FLAGS.output_dir)

  json_to_write = os.path.join(FLAGS.output_dir, 'output%d.json' % host_id)
  all_jsons = []

  # In a case of a reset we have to check how much work was already done.
  # We can check whether the processing of an example was finished, but
  # currently we are only checking whether it was started.
  done = FLAGS.starting_example
  reload_count = 0
  all_existing_files = tf.io.gfile.listdir(FLAGS.output_dir)
  for filename in all_existing_files:
    if 'processing' in filename:
      # The definition of digits looks for a number after the infix "processing"
      # in the file name. Example: tom_processing_532 will lead to
      # digits = "processing_532" and number equal to "532".
      digits = filename[filename.find('processing'):]
      number = ''.join(d for d in digits if d.isdigit())
      if is_number(
          number) and int(number) < FLAGS.num_examples + FLAGS.starting_example:
        done = max(done, int(number))
  print('The done number is {}'.format(done))

  if FLAGS.use_eval_set:
    drop_gen = trax_data.CreateDropInputs(train=False)()
  else:
    drop_gen = trax_data.CreateDropInputs(train=True)()
  padding_fun = trax_data.PadToLength()

  # TODO(henrykm): improve managment of the counters.
  # example_count_total - all numeric examples
  # example_count - all numeric examples above starting_example
  # reload_count - if we processed FLAGS.reload_after examples,
  #   then the checkpoint should be reloaded.
  # idx - total number of exaples
  example_count_total = 0
  reload_count += 1
  for idx, e in enumerate(drop_gen):
    if reload_count >= FLAGS.reload_after:
      vocab, model, initial_state = prepare_model(model_file, FLAGS.batch_size)
      reload_count = 0
    if example_count >= FLAGS.num_examples:
      print('Reached the example_count {} - breaking'.format(example_count))
      break
    if not is_number(e[1]):
      continue
    target_answer = float(e[1])

    # We count numeric starting examples
    example_count_total += 1
    if example_count_total <= FLAGS.starting_example:
      print('Skipping example_count_total {} because it is below {}'.format(
          example_count_total, FLAGS.starting_example))
      continue

    if example_count % 10 == 0:
      elapsed_time = time.time() - start_time
      start_time = time.time()
      print('Starting inference on example %d, %.2fs since last log' %
            (example_count, elapsed_time), flush=True)

    example_count += 1
    if example_count <= done - FLAGS.starting_example + 1:
      print('Skipping example_count {} because it is below {}'.format(
          example_count, done - FLAGS.starting_example))
      # We are increasing the example_count because the example
      # was processed before
      continue

    if example_count % host_count != host_id:
      continue

    # At this point we are committed to the processing of an example with
    # index example_count
    processing_file = os.path.join(FLAGS.output_dir, 'processing_')
    data_id = str(example_count + FLAGS.starting_example)
    with tf.io.gfile.GFile(processing_file + data_id, 'w') as w:
      w.write('Procesing started.')
    for repetition_id, example in multiply_examples(e):
      question = example[0]
      question_text = question[question.find(':') + 2:]
      question_text = question_text.replace('-', ' - ')
      question = 'infer full calculation: ' + question_text

      list_num = [
          float(num.replace(',', '').rstrip('.')) for num in re.findall(
              r'[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?', question)
      ]
      for i in range(len(list_num)):
        question += ' n{} = {}'.format(i, list_num[i])

      # print('Question {}'.format(question))
      tokenized_question = next(
          padding_fun(
              trax_data.tokenize([
                  question,
              ],
                                 vocab_file=gin.query_parameter(
                                     'trax.data.Tokenize.vocab_file'))))
      state = model.state
      if FLAGS.use_beam_search:
        answer_beams = decoding.beam_search(
            model,
            tokenized_question[None, :],
            n_beams=FLAGS.n_beams,
            max_length=FLAGS.max_answer_len,
            accelerate=False)
        model.state = state
      else:
        answer_beams = []
        # We recycle the n_beams flag to control the number
        # of autoregressive samples.
        for i in range(FLAGS.n_beams):
          answer = decoding.autoregressive_sample(
              model,
              tokenized_question[None, :],
              temperature=FLAGS.autoregressive_sample_temp,
              max_length=FLAGS.max_answer_len,
              accelerate=False)
          model.state = state
          answer_beams.append(answer)

      correct_example_index = -1

      for i in range(len(answer_beams)):
        if FLAGS.use_beam_search:
          answer = trax_data.detokenize(
              answer_beams[i][0][0],
              vocab_file=gin.query_parameter('trax.data.Tokenize.vocab_file'))
        else:
          answer = trax_data.detokenize(
              answer_beams[i][0],
              vocab_file=gin.query_parameter('trax.data.Tokenize.vocab_file'))
        print('Proposed computation {}'.format(answer))
        list_op = answer.split('|')
        if not list_op[-1]:
          list_op = list_op[:-1]

        try:
          result = trax_data.tf_inputs.compute_result(list_op, list_num)
          if target_answer in result:
            correct_example_index = result.index(target_answer)
            break
        # This is a temporary hack with "broad" exceptions - the computations
        # must fail sometime, because we evaluate arbitrary sequences; I am in
        # the process of checking what are possible failure modes.
        except Exception as e:  # pylint: disable=broad-except
          print(e)
          try:
            result = trax_data.tf_inputs.compute_result(list_op[:-1], list_num)
            if target_answer in result:
              correct_example_index = result.index(target_answer)
              break
          except Exception as e:  # pylint: disable=broad-except
            print(e)
            print('Infered incorrect computation.')

      if correct_example_index == -1:
        continue

      json_record = {
          'question': question_text,
          'input': question,
          'calculation': '|'.join(list_op[:correct_example_index + 1]),
          'target_answer': target_answer
      }
      all_jsons.append(json.dumps(json_record) + '\n')
      # Outputting the inferred data in JSONL format.
      data_id = str(example_count + FLAGS.starting_example)
      with tf.io.gfile.GFile(json_to_write + data_id, 'w') as w:
        w.write(json.dumps(json_record) + '\n')
    with tf.io.gfile.GFile(processing_file + data_id, 'w') as w:
      w.write('Procesing finished.')

  with tf.io.gfile.GFile(json_to_write + '_' + str(FLAGS.starting_example),
                         'w') as w:
    for record in all_jsons:
      w.write(record)


if __name__ == '__main__':
  absl_app.run(main)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Implementations of common recurrent neural network cells (RNNs)."""

from trax import fastmath
from trax.fastmath import numpy as jnp
from trax.layers import activation_fns
from trax.layers import base
from trax.layers import combinators as cb
from trax.layers import convolution
from trax.layers import core
from trax.layers import initializers


class LSTMCell(base.Layer):
  """LSTM Cell.

  For a nice overview of the motivation and (i, o, f) gates, see this tutorial:
  https://colah.github.io/posts/2015-08-Understanding-LSTMs/

  See this paper for a description and detailed study of all gate types:
  https://arxiv.org/pdf/1503.04069.pdf
  """

  def __init__(self,
               n_units,
               forget_bias=1.0,
               kernel_initializer=initializers.GlorotUniformInitializer(),
               bias_initializer=initializers.RandomNormalInitializer(1e-6)):
    super().__init__(n_in=2, n_out=2)
    self._n_units = n_units
    self._forget_bias = forget_bias
    self._kernel_initializer = kernel_initializer
    self._bias_initializer = bias_initializer

  def forward(self, inputs):
    x, lstm_state = inputs

    # LSTM state consists of c and h.
    c, h = jnp.split(lstm_state, 2, axis=-1)

    # Dense layer on the concatenation of x and h.
    w, b = self.weights
    y = jnp.dot(jnp.concatenate([x, h], axis=-1), w) + b

    # i = input_gate, j = new_input, f = forget_gate, o = output_gate
    i, j, f, o = jnp.split(y, 4, axis=-1)

    new_c = c * fastmath.sigmoid(f) + fastmath.sigmoid(i) * jnp.tanh(j)
    new_h = jnp.tanh(new_c) * fastmath.sigmoid(o)
    return new_h, jnp.concatenate([new_c, new_h], axis=-1)

  def init_weights_and_state(self, input_signature):
    # LSTM state last dimension must be twice n_units.
    if input_signature[1].shape[-1] != 2 * self._n_units:
      raise ValueError(
          f'Last dimension of state (shape: {str(input_signature[1].shape)}) '
          f'must be equal to 2*n_units ({2 * self._n_units})')
    # The dense layer input is the input and half of the lstm state.
    input_shape = input_signature[0].shape[-1] + self._n_units
    rng1, rng2 = fastmath.random.split(self.rng, 2)
    w = self._kernel_initializer((input_shape, 4 * self._n_units), rng1)
    b = self._bias_initializer((4 * self._n_units,), rng2) + self._forget_bias
    self.weights = (w, b)


def MakeZeroState(depth_multiplier=1):
  """Makes zeros of shape like x but removing the length (axis 1)."""
  def f(x):  # pylint: disable=invalid-name
    if len(x.shape) != 3:
      raise ValueError(f'Layer input should be a rank 3 tensor representing'
                       f' (batch_size, sequence_length, feature_depth); '
                       f'instead got shape {x.shape}.')
    return jnp.zeros((x.shape[0], depth_multiplier * x.shape[-1]),
                     dtype=jnp.float32)
  return base.Fn('MakeZeroState', f)


def LSTM(n_units, mode='train', return_state=False, initial_state=False):
  """LSTM running on axis 1.

  Args:
    n_units: `n_units` for the `LSTMCell`.
    mode: if 'predict' then we save the previous state for one-by-one inference.
    return_state: Boolean. Whether to return the latest status in addition to
      the output. Default: False.
    initial_state: Boolean. If the state RNN (c, h) is to be obtained from the
      stack. Default: False.

  Returns:
    A LSTM layer.
  """

  if not initial_state:
    zero_state = MakeZeroState(depth_multiplier=2)  # pylint: disable=no-value-for-parameter
    if return_state:
      return cb.Serial(
          cb.Branch([], zero_state),
          cb.Scan(LSTMCell(n_units=n_units), axis=1, mode=mode),
          name=f'LSTM_{n_units}',
          sublayers_to_print=[])
    else:
      return cb.Serial(
          cb.Branch([], zero_state),  # fill state RNN with zero.
          cb.Scan(LSTMCell(n_units=n_units), axis=1, mode=mode),
          cb.Select([0], n_in=2),  # Drop RNN state.
          # Set the name to LSTM and don't print sublayers.
          name=f'LSTM_{n_units}', sublayers_to_print=[])
  else:
    if return_state:
      return cb.Serial(
          cb.Scan(LSTMCell(n_units=n_units), axis=1, mode=mode),
          name=f'LSTM_{n_units}', sublayers_to_print=[])
    else:
      return cb.Serial(
          cb.Scan(LSTMCell(n_units=n_units), axis=1, mode=mode),
          cb.Select([0], n_in=2),  # Drop RNN state.
          name=f'LSTM_{n_units}', sublayers_to_print=[])


class GRUCell(base.Layer):
  """Builds a traditional GRU cell with dense internal transformations.

  Gated Recurrent Unit paper: https://arxiv.org/abs/1412.3555
  """

  def __init__(self,
               n_units,
               forget_bias=0.0,
               kernel_initializer=initializers.RandomUniformInitializer(0.01),
               bias_initializer=initializers.RandomNormalInitializer(1e-6)):
    super().__init__(n_in=2, n_out=2)
    self._n_units = n_units
    self._forget_bias = forget_bias
    self._kernel_initializer = kernel_initializer
    self._bias_initializer = bias_initializer

  def forward(self, inputs):
    x, gru_state = inputs

    # Dense layer on the concatenation of x and h.
    w1, b1, w2, b2 = self.weights
    y = jnp.dot(jnp.concatenate([x, gru_state], axis=-1), w1) + b1

    # Update and reset gates.
    u, r = jnp.split(fastmath.sigmoid(y), 2, axis=-1)

    # Candidate.
    c = jnp.dot(jnp.concatenate([x, r * gru_state], axis=-1), w2) + b2

    new_gru_state = u * gru_state + (1 - u) * jnp.tanh(c)
    return new_gru_state, new_gru_state

  def init_weights_and_state(self, input_signature):
    if input_signature[1].shape[-1] != self._n_units:
      raise ValueError(
          f'Second argument in input signature should have a final dimension of'
          f' {self._n_units}; instead got {input_signature[1].shape[-1]}.')

    # The dense layer input is the input and half of the GRU state.
    input_shape = input_signature[0].shape[-1] + self._n_units
    rng1, rng2, rng3, rng4 = fastmath.random.split(self.rng, 4)
    w1 = self._kernel_initializer((input_shape, 2 * self._n_units), rng1)
    b1 = self._bias_initializer((2 * self._n_units,), rng2) + self._forget_bias
    w2 = self._kernel_initializer((input_shape, self._n_units), rng3)
    b2 = self._bias_initializer((self._n_units,), rng4)
    self.weights = (w1, b1, w2, b2)


def GRU(n_units, mode='train'):
  """GRU running on axis 1."""
  zero_state = MakeZeroState(depth_multiplier=1)  # pylint: disable=no-value-for-parameter
  return cb.Serial(
      cb.Branch([], zero_state),
      cb.Scan(GRUCell(n_units=n_units), axis=1, mode=mode),
      cb.Select([0], n_in=2),  # Drop RNN state.
      # Set the name to GRU and don't print sublayers.
      name=f'GRU_{n_units}', sublayers_to_print=[]
  )


def ConvGRUCell(n_units, kernel_size=(3, 3)):
  """Builds a convolutional GRU.

  Paper: https://arxiv.org/abs/1511.06432.

  Args:
    n_units: Number of hidden units
    kernel_size: Kernel size for convolution

  Returns:
    A Stax model representing a GRU cell with convolution transforms.
  """

  def BuildConv():
    return convolution.Conv(
        filters=n_units, kernel_size=kernel_size, padding='SAME')

  return GeneralGRUCell(
      candidate_transform=BuildConv,
      memory_transform_fn=None,
      gate_nonlinearity=activation_fns.Sigmoid,
      candidate_nonlinearity=activation_fns.Tanh)


def GeneralGRUCell(candidate_transform,
                   memory_transform_fn=None,
                   gate_nonlinearity=activation_fns.Sigmoid,
                   candidate_nonlinearity=activation_fns.Tanh,
                   dropout_rate_c=0.1,
                   sigmoid_bias=0.5):
  r"""Parametrized Gated Recurrent Unit (GRU) cell construction.

  GRU update equations for update gate, reset gate, candidate memory, and new
  state:

  .. math::
    u_t &= \sigma(U' \times s_{t-1} + B') \\
    r_t &= \sigma(U'' \times s_{t-1} + B'') \\
    c_t &= \tanh(U \times (r_t \odot s_{t-1}) + B) \\
    s_t &= u_t \odot s_{t-1} + (1 - u_t) \odot c_t

  See `combinators.Gate` for details on the gating function.


  Args:
    candidate_transform: Transform to apply inside the Candidate branch. Applied
      before nonlinearities.
    memory_transform_fn: Optional transformation on the memory before gating.
    gate_nonlinearity: Function to use as gate activation; allows trying
      alternatives to `Sigmoid`, such as `HardSigmoid`.
    candidate_nonlinearity: Nonlinearity to apply after candidate branch; allows
      trying alternatives to traditional `Tanh`, such as `HardTanh`.
    dropout_rate_c: Amount of dropout on the transform (c) gate. Dropout works
      best in a GRU when applied exclusively to this branch.
    sigmoid_bias: Constant to add before sigmoid gates. Generally want to start
      off with a positive bias.

  Returns:
    A model representing a GRU cell with specified transforms.
  """
  gate_block = [  # u_t
      candidate_transform(),
      _AddSigmoidBias(sigmoid_bias),
      gate_nonlinearity(),
  ]
  reset_block = [  # r_t
      candidate_transform(),
      _AddSigmoidBias(sigmoid_bias),  # Want bias to start positive.
      gate_nonlinearity(),
  ]
  candidate_block = [
      cb.Dup(),
      reset_block,
      cb.Multiply(),  # Gate S{t-1} with sigmoid(candidate_transform(S{t-1}))
      candidate_transform(),  # Final projection + tanh to get Ct
      candidate_nonlinearity(),  # Candidate gate

      # Only apply dropout on the C gate. Paper reports 0.1 as a good default.
      core.Dropout(rate=dropout_rate_c)
  ]
  memory_transform = memory_transform_fn() if memory_transform_fn else []
  return cb.Serial(
      cb.Branch(memory_transform, gate_block, candidate_block),
      cb.Gate(),
  )


def InnerSRUCell():
  """The inner (non-parallel) computation of an SRU."""
  def f(cur_x_times_one_minus_f, cur_f, cur_state):  # pylint: disable=invalid-name
    res = cur_f * cur_state + cur_x_times_one_minus_f
    return res, res
  return base.Fn('InnerSRUCell', f, n_out=2)


def ScanSRUCell(mode, monkey_patched_mask=None):
  """The inner (non-parallel) computation of an SRU."""
  if monkey_patched_mask is None:
    return cb.Scan(InnerSRUCell(), axis=1, mode=mode)

  # This is necessary for Terraformer model. See comments there.
  # The mask will only be used in Terraformer in predict mode.
  assert mode == 'predict'

  def update_mask(mask, x_times_one_minus_f):  # pylint: disable=invalid-name
    initial = jnp.ones(x_times_one_minus_f.shape[:2], dtype=jnp.float32)
    if initial.shape[1] > 1:
      updated_mask = fastmath.dynamic_update_slice_in_dim(
          initial != 0, mask != 0, 1, axis=1)
    else:
      updated_mask = initial
    return updated_mask, x_times_one_minus_f

  def masked_inner_sru_cell(cur_mask, cur_x_times_one_minus_f, cur_f,  # pylint: disable=invalid-name
                            cur_state):
    res = ((cur_f * cur_state + cur_x_times_one_minus_f) * cur_mask
           + (1 - cur_mask) * cur_state)
    return res, res

  return cb.Serial(
      monkey_patched_mask.get_layer(),
      base.Fn('update_mask', update_mask, n_out=2),
      cb.Scan(base.Fn('MaskedInnerSRUCell', masked_inner_sru_cell, n_out=2),
              axis=1, mode=mode),
      )


def SRU(n_units, activation=None, mode='train'):
  r"""SRU (Simple Recurrent Unit) layer as in https://arxiv.org/abs/1709.02755.

  As defined in the paper:

  .. math::
    y_t &= W x_t + B \quad \hbox{(include $B$ optionally)} \\
    f_t &= \sigma(Wf x_t + bf) \\
    r_t &= \sigma(Wr x_t + br) \\
    c_t &= f_t \times c_{t-1} + (1 - f_t) \times y_t \\
    h_t &= r_t \times \hbox{activation}(c_t) + (1 - r_t) \times x_t

  We assume the input is of shape [batch, length, depth] and recurrence
  happens on the length dimension. This returns a single layer. It's best
  to use at least 2, they say in the paper, except inside a Transformer.

  Args:
    n_units: output depth of the SRU layer.
    activation: Optional activation function.
    mode: if 'predict' then we save the previous state for one-by-one inference

  Returns:
    The SRU layer.
  """
  sigmoid_activation = activation_fns.Sigmoid()
  return cb.Serial(                                         # x
      cb.Branch(core.Dense(3 * n_units), []),               # r_f_y, x
      cb.Split(n_items=3),                                  # r, f, y, x
      cb.Parallel(sigmoid_activation, sigmoid_activation),  # r, f, y, x
      base.Fn('',
              lambda r, f, y: (y * (1.0 - f), f, r),    # y * (1 - f), f, r, x
              n_out=3),
      cb.Parallel([], [], cb.Branch(MakeZeroState(), [])),
      ScanSRUCell(mode=mode),
      cb.Select([0], n_in=2),                               # act(c), r, x
      activation if activation is not None else [],
      base.Fn('FinalSRUGate', lambda c, r, x: c * r + x * (1 - r) * (3**0.5)),
      # Set the name to SRU and don't print sublayers.
      name=f'SRU_{n_units}', sublayers_to_print=[]
  )


def _AddSigmoidBias(sigmoid_bias):
  return base.Fn('AddSigmoidBias({sigmoid_bias})',
                 lambda x: x + sigmoid_bias)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""RL environments created from supervised data-sets."""

import gym
import numpy as np


class SequenceDataEnv(object):
  """RL environment created from a generator of sequential data.

  This class allows to create RL environments from supervised sequential data,
  such as tokenized natural languague processing tasks. The data comes as:
    (input1, output1, input2, output2, ...)
  where inputs and outputs are all sequences of integers.

  For example, with input (2, 3) and output (4, 5), so data = [(2, 3), (4, 5)],
  the sequence of (observations, rewards, actions) will look like:
    2                   = env.reset()                # first observation
    3,    0.0,   _, _   = env.step(ignored_action)
    eos,  0.0,   _, _   = env.step(ignored_action)
    act1, 0.0,   _, _   = env.step(act1)             # observation = action
    act2, 0.0,   _, _   = env.step(act2)             # observation = action
    eos,  score, _, _   = env.step(eos)

  where score = metric((4, 5), (act1, act2)) is the reward gotten from
  comparing the two actions to the actual output from the data.

  The environment first presents the input as observations, doing this
  sequentially, token-by-token, and ignoring all actions taken by the policy.
  Then, the policy is asked to generate the response, again, token-by-token,
  until it generates EOS. Generated tokens are repeated as observations.
  When EOS is encountered, a metric is computed between the generated
  output and the output from data, and this metric is returned as reward.
  """

  def __init__(self, data_stream, vocab_size, metric=None,
               eos_id=1, max_length=1000):
    """The constructor.

    Args:
      data_stream: A python generator creating lists or tuples of
        sequences (list, tuples or numpy arrays) of integers.
      vocab_size: Integer, the size of the vocabulary. All integers in the
        data stream must be positive and smaller than this value.
      metric: A function taking two lists of integers and returning a float.
        If None, we use per-token accuracy as the default metric.
      eos_id: Integer, the id of the EOS symbol.
      max_length: Integer, maximum length of the policy reply to avoid
        infinite episodes if policy never produces EOS.

    Returns:
      A new environment which presents the data and compares the policy
      response with the expected data, returning metric as reward.
    """
    self._data = data_stream
    self._vocab_size = vocab_size
    self._eos = eos_id
    self._max_length = max_length
    self._metric = _accuracy if metric is None else metric
    self.reset()

  @property
  def _on_input(self):
    """Return True if we're currently processing input, False if output."""
    cur_sequence_id, _ = self._cur_position
    return cur_sequence_id % 2 == 0

  @property
  def observation(self):
    cur_sequence_id, cur_token_id = self._cur_position
    if cur_sequence_id >= len(self._cur_sequence):
      obs = self._eos
    elif self._on_input:
      obs = self._cur_sequence[cur_sequence_id][cur_token_id]
    else:
      obs = self._response[-1] if self._response else self._eos
    return np.array(int(obs), dtype=np.int32)

  @property
  def action_space(self):
    return gym.spaces.Discrete(self._vocab_size)

  @property
  def observation_space(self):
    return gym.spaces.Discrete(self._vocab_size)

  def reset(self):
    """Reset this environment."""
    self._cur_sequence = next(self._data)
    # Position contains 2 indices: which sequnece are we in? (input1, output1,
    # input2, output2 and so on) and which token in the sequence are we in?
    self._cur_position = (0, 0)
    self._response = []
    return self.observation

  def step(self, action):
    """Single step of the environment when policy took `action`."""
    cur_sequence_id, cur_token_id = self._cur_position
    if cur_sequence_id >= len(self._cur_sequence):
      return np.array(self._eos, dtype=np.int32), 0.0, True, None

    # Emit the control mask on the output.
    control_mask = int(not self._on_input)

    if self._on_input:
      self._response = []
      if cur_token_id + 1 < len(self._cur_sequence[cur_sequence_id]):
        self._cur_position = (cur_sequence_id, cur_token_id + 1)
        done = False
      else:
        self._cur_position = (cur_sequence_id + 1, 0)
        done = cur_sequence_id + 1 >= len(self._cur_sequence)
      reward = 0.0
      discount_mask = 0

    else:
      self._response.append(action)
      if action == self._eos or len(self._response) > self._max_length:
        self._cur_position = (cur_sequence_id + 1, 0)
        reward = self._metric(
            self._response[:-1], self._cur_sequence[cur_sequence_id])
        done = cur_sequence_id + 1 >= len(self._cur_sequence)
        # Emit the discount mask on the last token of each action.
        discount_mask = 1
      else:
        reward = 0.0
        done = False
        discount_mask = 0

    info = {'control_mask': control_mask, 'discount_mask': discount_mask}
    return self.observation, reward, done, info


def copy_stream(length, low=2, high=15, n=1):
  """Generate `n` random sequences of length `length` and yield with copies."""
  while True:
    res = []
    for _ in range(n):
      seq = np.random.randint(low, high, size=(length,), dtype=np.int32)
      res.extend([seq, seq])
    yield res


def _accuracy(seq1, seq2):
  """Token-level accuracy."""
  seq1, seq2 = np.array(seq1), np.array(seq2)
  max_length = max(seq1.shape[-1], seq2.shape[-1])
  min_length = min(seq1.shape[-1], seq2.shape[-1])
  seq1s, seq2s = seq1[..., :min_length], seq2[..., :min_length]
  return np.sum(np.equal(seq1s, seq2s)) / max_length




# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""BERT."""

import os
import pathlib
import urllib
import zipfile

import gin
import jax
import tensorflow as tf

from trax import fastmath
from trax import layers as tl
from trax.fastmath import numpy as np

# pylint: disable=invalid-name

_MODEL_LINKS = {
    'bert-base-uncased':
        'https://storage.googleapis.com/bert_models/2018_10_18/uncased_L-12_H-768_A-12.zip'
}


class AddBias(tl.Layer):

  def forward(self, inputs):
    x = inputs
    return x + self.weights

  def init_weights_and_state(self, input_signature):
    self.weights = np.zeros(input_signature.shape[-1])


@gin.configurable
def BERTClassifierHead(n_classes):
  return tl.Serial([
      tl.Select([0], n_in=2),
      tl.Dense(
          n_classes,
          kernel_initializer=tl.RandomNormalInitializer(0.02),
          bias_initializer=tl.RandomNormalInitializer(1e-6),
      ),
  ])


@gin.configurable
def BERTRegressionHead():
  return tl.Serial([
      tl.Select([0], n_in=2),
      tl.Dense(
          1,
          kernel_initializer=tl.RandomNormalInitializer(0.02),
          bias_initializer=tl.RandomNormalInitializer(1e-6),
      ),
      tl.Fn('RemoveAxis', lambda x: np.squeeze(x, axis=1))
  ])


@gin.configurable
def BERTMLMHead(vocab_size=30522):
  return tl.Serial([
      tl.Select([1], n_in=2),
      tl.Dense(vocab_size,
               kernel_initializer=tl.RandomNormalInitializer(0.02),
               bias_initializer=tl.RandomNormalInitializer(1e-6),
               ),
  ])


@gin.configurable
def BERTPretrainingLoss():
  nsp_loss = [tl.Select([0, 2, 3], n_in=6), tl.WeightedCategoryCrossEntropy()]
  mlm_loss = [tl.Select([1, 4, 5], n_in=6), tl.WeightedCategoryCrossEntropy()]
  return tl.Serial(tl.Branch(nsp_loss, mlm_loss), tl.Add())


@gin.configurable
def BERTPretrainingHead(n_classes):
  return tl.Branch(BERTClassifierHead(n_classes), BERTMLMHead()),


# TODO(piotrekp1): add tests


def BERT(
    d_model=768,
    vocab_size=30522,
    max_len=512,
    type_vocab_size=2,
    n_heads=12,
    d_ff=3072,
    n_layers=12,
    head=None,
    init_checkpoint=None,
    mode='eval',
):
  """BERT (default hparams are for bert-base-uncased)."""
  # TODO(piotrekp1): loading config from model_name

  layer_norm_eps = 1e-12
  d_head = d_model // n_heads

  word_embeddings = tl.Embedding(vocab_size, d_model)
  type_embeddings = tl.Embedding(type_vocab_size, d_model)
  position_embeddings = tl.PositionalEncoding(max_len, mode=mode)
  embeddings = [
      tl.Select([0, 1, 0], n_in=3),  # Drops 'idx' input.
      tl.Parallel(word_embeddings, type_embeddings, [
          tl.PaddingMask(),
          tl.Fn('Squeeze', lambda x: np.squeeze(x, (1, 2)), n_out=1)
      ]),
      tl.Add(),
      position_embeddings,
      tl.LayerNorm(epsilon=layer_norm_eps),
  ]

  encoder = []
  for _ in range(n_layers):
    attn = tl.SelfAttention(
        n_heads=n_heads,
        d_qk=d_head,
        d_v=d_head,
        bias=True,
        masked=True,
        mode=mode)
    feed_forward = [tl.Dense(d_ff), tl.Gelu(), tl.Dense(d_model)]
    encoder += [
        tl.Select([0, 1, 1]),  # Save a copy of the mask
        tl.Residual(attn, AddBias()),  # pylint: disable=no-value-for-parameter
        tl.LayerNorm(epsilon=layer_norm_eps),
        tl.Residual(*feed_forward),
        tl.LayerNorm(epsilon=layer_norm_eps),
    ]

  encoder += [tl.Select([0], n_in=2)]  # Drop the mask

  pooler = [
      tl.Fn('', lambda x: (x[:, 0, :], x), n_out=2),
      tl.Dense(d_model),
      tl.Tanh(),
  ]

  init_checkpoint = init_checkpoint if mode == 'train' else None
  bert = PretrainedBERT(
      embeddings + encoder + pooler, init_checkpoint=init_checkpoint)

  if head is not None:
    bert = tl.Serial(bert, head())

  return bert


class PretrainedBERT(tl.Serial):
  """Wrapper that always initializes weights from a pre-trained checkpoint."""

  def __init__(self, *sublayers, init_checkpoint=None):
    super().__init__(*sublayers)


    if init_checkpoint is None:
      # initialize model from scratch
      self.init_checkpoint = None
    elif os.path.sep not in init_checkpoint:
      # initialize model from model name
      init_checkpoint_dir, init_checkpoint_filename = self.download_model(
          init_checkpoint)
      self.init_checkpoint = os.path.join(init_checkpoint_dir,
                                          init_checkpoint_filename)
    else:
      # initialize model from path
      self.init_checkpoint = init_checkpoint

  # If init_checkpoint is a model name and there is no local
  # model with that name then it downloads it and returns newly created path.
  @classmethod
  def download_model(cls, model_name):
    """Returns model dir path with model filename."""
    try:
      model_link = _MODEL_LINKS[model_name]
    except KeyError:
      raise KeyError(
          f'Not known model name, please make sure the model name'
          f' is in the list of available models. If this is a path'
          f' to a model it should contain at least one {os.path.sep}')

    def find_ckpt_file_in_dir(dir_path):
      # TODO(piotrekp1): get last checkpoint instead of any
      for fname in os.listdir(dir_path):
        if '.ckpt' in fname:
          return fname.split('.ckpt')[0] + '.ckpt'
      raise FileNotFoundError('Selected directory doesn\'t contain a ckpt file')

    def cd_nested_directory(dir_path):
      while len(os.listdir(dir_path)) == 1:
        filename = os.listdir(dir_path)[0]
        new_path = os.path.join(dir_path, filename)
        if os.path.isdir(new_path):
          dir_path = new_path
      return dir_path

    download_dir = os.path.join('~', 'trax', model_name)
    download_dir = os.path.expanduser(download_dir)

    if os.path.exists(download_dir):
      if os.path.isdir(download_dir) and len(os.listdir(download_dir)):
        # model already exists, mdoel directory as input
        download_dir = cd_nested_directory(
            download_dir)  # go in if nested single directories inside
        return download_dir, find_ckpt_file_in_dir(download_dir)
    with pathlib.Path(download_dir) as p:
      p.mkdir(parents=True, exist_ok=True)

    if not model_link.endswith('.zip'):
      raise NotImplementedError(
          'Only downloading models packed with zip is implemented')

    # assumes model is packed with zip
    download_path = os.path.join(download_dir, 'model_temp')

    print(f'Downloading model from {model_link} to {download_dir}'
         )  # TODO(piotrekp1) logging to stderr?
    urllib.request.urlretrieve(model_link, download_path)

    if not zipfile.is_zipfile(download_path):
      raise NotImplementedError(
          'Only downloading models packed with zip is implemented')

    # TODO(piotrekp1): some locks to handle crashes during unpacking or download
    with zipfile.ZipFile(download_path, 'r') as zip_ref:
      zip_ref.extractall(download_dir)

    os.remove(download_path)
    download_dir = cd_nested_directory(download_dir)

    return download_dir, find_ckpt_file_in_dir(download_dir)

  def init_weights_and_state(self, input_signature):
    super().init_weights_and_state(input_signature)
    if self.init_checkpoint is None:
      return

    print('Loading pre-trained weights from', self.init_checkpoint)
    ckpt = tf.train.load_checkpoint(self.init_checkpoint)

    def reshape_qkv(name):
      x = ckpt.get_tensor(name)
      return x.reshape((x.shape[0], -1, 64)).swapaxes(0, 1)

    def reshape_o(name):
      x = ckpt.get_tensor(name)
      return x.reshape((-1, 64, x.shape[-1]))

    def reshape_bias(name):
      x = ckpt.get_tensor(name)
      return x.reshape((-1, 64))

    new_w = [
        ckpt.get_tensor('bert/embeddings/word_embeddings'),
        ckpt.get_tensor('bert/embeddings/token_type_embeddings'),
        ckpt.get_tensor('bert/embeddings/position_embeddings')[None, ...],
        ckpt.get_tensor('bert/embeddings/LayerNorm/gamma'),
        ckpt.get_tensor('bert/embeddings/LayerNorm/beta'),
    ]

    for i in range(12):  # 12 layers
      new_w += [
          reshape_qkv(f'bert/encoder/layer_{i}/attention/self/query/kernel'),
          reshape_qkv(f'bert/encoder/layer_{i}/attention/self/key/kernel'),
          reshape_qkv(f'bert/encoder/layer_{i}/attention/self/value/kernel'),
          reshape_o(f'bert/encoder/layer_{i}/attention/output/dense/kernel'),
          reshape_bias(f'bert/encoder/layer_{i}/attention/self/query/bias'),
          reshape_bias(f'bert/encoder/layer_{i}/attention/self/key/bias'),
          reshape_bias(f'bert/encoder/layer_{i}/attention/self/value/bias'),
          ckpt.get_tensor(
              f'bert/encoder/layer_{i}/attention/output/dense/bias'),
          ckpt.get_tensor(
              f'bert/encoder/layer_{i}/attention/output/LayerNorm/gamma'),
          ckpt.get_tensor(
              f'bert/encoder/layer_{i}/attention/output/LayerNorm/beta'),
          ckpt.get_tensor(f'bert/encoder/layer_{i}/intermediate/dense/kernel'),
          ckpt.get_tensor(f'bert/encoder/layer_{i}/intermediate/dense/bias'),
          ckpt.get_tensor(f'bert/encoder/layer_{i}/output/dense/kernel'),
          ckpt.get_tensor(f'bert/encoder/layer_{i}/output/dense/bias'),
          ckpt.get_tensor(f'bert/encoder/layer_{i}/output/LayerNorm/gamma'),
          ckpt.get_tensor(f'bert/encoder/layer_{i}/output/LayerNorm/beta'),
      ]

    new_w += [
        ckpt.get_tensor('bert/pooler/dense/kernel'),
        ckpt.get_tensor('bert/pooler/dense/bias'),
    ]

    for a, b in zip(fastmath.tree_leaves(self.weights), new_w):
      assert a.shape == b.shape, (
          f'Expected shape {a.shape}, got shape {b.shape}')
    self.weights = jax.tree_unflatten(jax.tree_structure(self.weights), new_w)
    move_to_device = jax.jit(lambda x: x)
    self.weights = jax.tree_map(move_to_device, self.weights)

  def _settable_attrs(self):
    """We allow to set attributes required for loading the model from its checkpoints."""
    return (*super()._settable_attrs(), 'init_checkpoint')



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for rnn layers."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from trax import fastmath
from trax import shapes
import trax.layers as tl


BACKENDS = [fastmath.Backend.JAX]


@parameterized.named_parameters(
    ('_' + b.value, b) for b in BACKENDS)
class RnnTest(parameterized.TestCase):

  def test_conv_gru_cell(self, backend):
    with fastmath.use_backend(backend):
      layer = tl.ConvGRUCell(9, kernel_size=(3, 3))
      x = np.ones((8, 1, 7, 9))
      _, _ = layer.init(shapes.signature(x))
      y = layer(x)
      self.assertEqual(y.shape, x.shape)

  def test_gru_cell(self, backend):
    with fastmath.use_backend(backend):
      layer = tl.GRUCell(9)
      xs = [np.ones((8, 7, 9)), np.ones((8, 7, 9))]
      _, _ = layer.init(shapes.signature(xs))
      ys = layer(xs)
      self.assertEqual([y.shape for y in ys], [(8, 7, 9), (8, 7, 9)])

  def test_lstm_cell(self, backend):
    with fastmath.use_backend(backend):
      layer = tl.LSTMCell(9)
      xs = [np.ones((8, 9)), np.ones((8, 18))]
      _, _ = layer.init(shapes.signature(xs))
      ys = layer(xs)
      self.assertEqual([y.shape for y in ys], [(8, 9), (8, 18)])

  def test_sru(self, backend):
    with fastmath.use_backend(backend):
      layer = tl.SRU(7)
      x = np.ones((8, 9, 7), np.float32)
      _, _ = layer.init(shapes.signature(x))
      y = layer(x)
      self.assertEqual(y.shape, x.shape)

  def test_names(self, backend):
    with fastmath.use_backend(backend):
      layer = tl.LSTM(3)
      self.assertEqual('LSTM_3', str(layer))
      layer = tl.GRU(5)
      self.assertEqual('GRU_5', str(layer))
      layer = tl.SRU(7)
      self.assertEqual('SRU_7', str(layer))


if __name__ == '__main__':
  absltest.main()
